import requests
import time
import logging

logger = logging.getLogger(__name__)


class PandoraStatsReader(object):
    # TODO: maybe make stats collection asyncronous

    def next(self):
        try:
            pandora_response = requests.get("http://localhost:1234/debug/vars")
            pandora_stat = pandora_response.json()

            return [{
                'ts': int(time.time() - 1),
                'metrics': {
                    'instances': pandora_stat.get("engine_ActiveRequests"),
                    'reqps': pandora_stat.get("engine_ReqPS"),
                }
            }]
        except requests.ConnectionError:
            logger.info("Pandora expvar http interface is unavailable")
        except requests.HTTPError:
            logger.warning(
                "Pandora expvar http interface is unavailable", exc_info=True)
        except Exception:
            logger.warning(
                "Couldn't decode pandora stat:\n%s\n",
                pandora_response.text,
                exc_info=True)

        return [{
            'ts': int(time.time() - 1),
            'metrics': {
                'instances': 0,
                'reqps': 0
            }
        }]

    def close(self):
        pass

    def __iter__(self):
        return self



import os.path
import seaborn

from ...common.interfaces import AbstractPlugin, MonitoringDataListener, AggregateResultListener

import matplotlib
matplotlib.use("Agg")
import matplotlib.pyplot as plt  # noqa:E402

_ALL_ = "All"
_CHARTSETS = {
    "cpu-cpu-": {
        "CPU": _ALL_
    },
    "net-": {
        "Network": {"bytes_sent", "bytes_recv"}
    },
    "diskio-": {
        "Disk IO": {"read_bytes", "write_bytes"},
        "Disk latency": {"read_time", "write_time"}
    },
}
_CUSTOM_PREFIX = "custom:"
_REPORT_FILE_OPTION = "report_file"


class Plugin(AbstractPlugin, AggregateResultListener, MonitoringDataListener):
    """
        Generates simple shooting report in svg format
    """

    SECTION = "svgreport"

    def __init__(self, core):
        super(Plugin, self).__init__(core)

    def get_available_options(self):
        return [_REPORT_FILE_OPTION]

    def configure(self):
        self.__report_path = self.get_option(_REPORT_FILE_OPTION, "report.svg")
        if os.path.split(self.__report_path)[0] or os.path.splitdrive(
                self.__report_path)[0]:
            raise Exception("Only simple file names supported")

        self.__shooting_data = []
        self.__monitoring_data = []
        self.core.job.subscribe_plugin(self)

    def on_aggregated_data(self, data, stats):
        if data:
            self.__shooting_data.append(data)

    def monitoring_data(self, data_list):
        if data_list:
            self.__monitoring_data.extend(data_list)

    def post_process(self, retcode):
        monitoring_chartsets = self.__get_monitoring_chartsets()
        min_x = self.__shooting_data[0][
            "ts"]  # sync start of shooting and start of monitoring

        seaborn.set(style="whitegrid", palette="Set2")
        seaborn.despine()

        plot_count = len(monitoring_chartsets) + 1
        plt.figure(figsize=(16, 3 * plot_count))

        # testing
        plt.subplot(plot_count, 1, 1)
        plt.title("RPS")
        x, y = self.__get_shooting_coords("proto_code", min_x)
        for variant in x:
            plt.plot(x[variant], y[variant], label=variant)
        plt.gca().legend(fontsize="x-small")

        # monitoring
        for plot_num, chartset_data in enumerate(
                sorted(monitoring_chartsets.iteritems()), 1):
            chartset_title, signals = chartset_data

            plt.subplot(plot_count, 1, plot_num + 1)
            plt.title(chartset_title)

            for signal_name, signal_suffix in signals:
                x, y = self.__get_monitoring_coords(signal_name, min_x)
                plt.plot(x, y, label=signal_suffix)
            plt.gca().legend(fontsize="x-small")

        plt.tight_layout()
        plt.suptitle("Shooting results", fontsize=16, fontweight="bold")
        plt.subplots_adjust(top=0.96)
        plt.savefig(os.path.join(self.core.artifacts_dir, self.__report_path))

    def __find_monitoring_chartset(self, signal_prefix, signal_suffix):
        """
            Tune chartset content

            Some of signals should be skipped, and other should be distributed between
            two chartsets
        """

        if signal_prefix.startswith(_CUSTOM_PREFIX):
            signal_prefix = signal_prefix[len(_CUSTOM_PREFIX):]

        for chartset_prefix, chartset_data in _CHARTSETS.iteritems():
            if signal_prefix.startswith(chartset_prefix):
                for chartset_title, chartset_signals in chartset_data.iteritems(
                ):
                    if chartset_signals is _ALL_ or signal_suffix in chartset_signals:
                        return "{} {}".format(
                            chartset_title,
                            signal_prefix[len(chartset_prefix):])
                else:
                    return None
        else:
            return signal_prefix

    def __get_monitoring_chartsets(self):
        """Analyze monitoring signals and organize chartsets"""

        chartsets = {}
        for p in self.__monitoring_data:
            metrics = p["data"]["localhost"]["metrics"]
            for signal_name, signal_value in metrics.iteritems():
                if not signal_value:
                    continue

                signal_prefix, signal_suffix = signal_name.split("_", 1)
                chartset_title = self.__find_monitoring_chartset(
                    signal_prefix, signal_suffix)
                if not chartset_title:
                    continue

                chartsets.setdefault((chartset_title), set()).add(
                    (signal_name, signal_suffix))

        return chartsets

    def __get_monitoring_coords(self, signal_name, min_x):
        """Return values for x and y axes of plot for specified signal"""

        x, y = [], []
        for p in self.__monitoring_data:
            metrics = p["data"]["localhost"]["metrics"]
            timestamp = p["timestamp"]
            if signal_name in metrics and timestamp >= min_x:
                x.append(timestamp - min_x)
                y.append(metrics[signal_name])
        return x, y

    def __get_shooting_coords(self, signal_name, min_x):
        """Return values for x and y axes of plot for specified signal"""

        x = {}
        y = {}
        for data in self.__shooting_data:
            timestamp = data["ts"]
            for variant, count in data["overall"][signal_name][
                    "count"].iteritems():
                x.setdefault(variant, []).append(timestamp - min_x)
                y.setdefault(variant, []).append(count)
        return x, y



from __future__ import division, absolute_import

import datetime
import time

from builtins import super
from ...common.interfaces import AbstractInfoWidget

from ..Console import screen as ConsoleScreen


class MavenInfoWidget(AbstractInfoWidget):
    ''' Right panel widget '''

    def __init__(self, owner):
        # FIXME python version 2.7 does not support this syntax. super() should
        # have arguments in Python 2
        super().__init__()
        self.krutilka = ConsoleScreen.krutilka()
        self.owner = owner

    def get_index(self):
        return 0

    def on_aggregated_data(self, data, stats):
        pass

    def render(self, screen):
        text = " Maven Test %s" % next(self.krutilka)
        space = screen.right_panel_width - len(text) - 1
        left_spaces = space // 2
        right_spaces = space // 2

        dur_seconds = int(time.time()) - int(self.owner.process_start_time)
        duration = str(datetime.timedelta(seconds=dur_seconds))

        template = screen.markup.BG_BROWN + '~' * left_spaces + \
            text + ' ' + '~' * right_spaces + screen.markup.RESET + "\n"
        template += "Command Line: %s\n"
        template += "    Duration: %s"
        data = (self.owner.maven_cmd, duration)

        return template % data



import logging
import time

import pip
from ...common.interfaces import AbstractPlugin, GeneratorPlugin

from .guns import LogGun, SqlGun, CustomGun, HttpGun, ScenarioGun, UltimateGun
from .reader import BfgReader, BfgStatsReader
from .widgets import BfgInfoWidget
from .worker import BFG
from ..Aggregator import Plugin as AggregatorPlugin
from ..Console import Plugin as ConsolePlugin
from ...stepper import StepperWrapper


class Plugin(AbstractPlugin, GeneratorPlugin):
    ''' Big Fucking Gun plugin '''
    SECTION = 'bfg'

    def __init__(self, core):
        self.log = logging.getLogger(__name__)
        AbstractPlugin.__init__(self, core)
        self.gun_type = None
        self.start_time = time.time()
        self.stepper_wrapper = StepperWrapper(self.core, Plugin.SECTION)
        self.log.info("Initialized BFG")

        self.gun_classes = {
            'log': LogGun,
            'sql': SqlGun,
            'custom': CustomGun,
            'http': HttpGun,
            'scenario': ScenarioGun,
            'ultimate': UltimateGun,
        }

    @staticmethod
    def get_key():
        return __file__

    def get_available_options(self):
        return [
            "gun_type", "instances", "cached_stpd", "pip"
        ] + self.stepper_wrapper.get_available_options

    def configure(self):
        self.log.info("Configuring BFG...")
        self.stepper_wrapper.read_config()

    def prepare_test(self):
        pip_deps = self.get_option("pip", "").splitlines()
        self.log.info("Installing with PIP: %s", pip_deps)
        if pip_deps:
            retcode = pip.main(["install", "--user"] + pip_deps)
            if retcode != 0:
                raise RuntimeError("Could not install required deps")
            import site
            reload(site)
        self.log.info("BFG using ammo type %s", self.get_option("ammo_type"))
        self.stepper_wrapper.prepare_stepper()
        gun_type = self.get_option("gun_type")
        if gun_type in self.gun_classes:
            self.gun = self.gun_classes[gun_type](self.core)
        else:
            raise NotImplementedError(
                'No such gun type implemented: "%s"' % gun_type)
        cached_stpd_option = self.get_option("cached_stpd", '0')
        if cached_stpd_option == '1':
            cached_stpd = True
        else:
            cached_stpd = False
        self.bfg = BFG(
            gun=self.gun,
            instances=self.stepper_wrapper.instances,
            stpd_filename=self.stepper_wrapper.stpd,
            cached_stpd=cached_stpd)
        aggregator = None
        try:
            aggregator = self.core.get_plugin_of_type(AggregatorPlugin)
        except Exception as ex:
            self.log.warning("No aggregator found: %s", ex)

        if aggregator:
            aggregator.reader = BfgReader(self.bfg.results)
            aggregator.stats_reader = BfgStatsReader(
                self.bfg.instance_counter, self.stepper_wrapper.steps)

        try:
            console = self.core.get_plugin_of_type(ConsolePlugin)
        except Exception as ex:
            self.log.debug("Console not found: %s", ex)
            console = None

        if console:
            widget = BfgInfoWidget()
            console.add_info_widget(widget)
            if aggregator:
                aggregator.add_result_listener(widget)
        self.log.info("Prepared BFG")

    def start_test(self):
        self.log.info("Starting BFG")
        self.start_time = time.time()
        self.bfg.start()

    def is_test_finished(self):
        if self.bfg.running():
            return -1
        else:
            self.log.info("BFG finished")
            return 0

    def end_test(self, retcode):
        if self.bfg.running():
            self.log.info("Terminating BFG")
            self.bfg.stop()
        return retcode



import logging


class AbstractPlugin(object):
    """ Plugin interface
    Parent class for all plugins """

    SECTION = 'DEFAULT'

    @staticmethod
    def get_key():
        """ Get dictionary key for plugin,
        should point to __file__ magic constant """
        raise TypeError("Abstract method needs to be overridden")

    def __init__(self, core):
        """

        @type core: TankCore
        """
        self.log = logging.getLogger(__name__)
        self.core = core

    def configure(self):
        """ A stage to read config values and instantiate objects """
        pass

    def prepare_test(self):
        """        Test preparation tasks        """
        pass

    def start_test(self):
        """        Launch test process        """
        pass

    def is_test_finished(self):
        """
        Polling call, if result differs from -1 then test end
        will be triggeted
        """
        return -1

    def end_test(self, retcode):
        """
        Stop processes launched at 'start_test',
        change return code if necessary
        """
        return retcode

    def post_process(self, retcode):
        """ Post-process test data """
        return retcode

    def get_option(self, option_name, default_value=None):
        """ Wrapper to get option from plugins' section """
        return self.core.get_option(self.SECTION, option_name, default_value)

    def set_option(self, option_name, value):
        """ Wrapper to set option to plugins' section """
        return self.core.set_option(self.SECTION, option_name, value)

    def get_available_options(self):
        """ returns array containing known options for plugin """
        return []

    def get_multiline_option(self, option_name, default_value=None):
        if default_value is not None:
            default = ' '.join(default_value)
        else:
            default = None
        value = self.get_option(option_name, default)
        if value:
            return (' '.join(value.split("\n"))).split(' ')
        else:
            return ()

    def publish(self, key, value):
        """publish value to status"""
        self.log.debug(
            "Publishing status: %s/%s: %s", self.__class__.__name__, key, value)
        self.core.publish(self.__class__.__name__, key, value)

    def close(self):
        """
        Release allocated resources here.
        Warning: don't do any logic or potentially dangerous operations
        """
        pass


class MonitoringDataListener(object):
    """ Monitoring interface
    parent class for Monitoring data listeners"""

    def __init__(self):
        pass

    def monitoring_data(self, data):
        """Notification about new monitoring data lines"""
        raise NotImplementedError("Abstract method needs to be overridden")


class AggregateResultListener(object):
    """ Listener interface
    parent class for Aggregate results listeners"""

    def on_aggregated_data(self, data, stats):
        """
        notification about new aggregated data and stats

        data contains aggregated metrics and stats contain non-aggregated
        metrics from gun (like instances count, for example)

        data and stats are cached and synchronized by timestamp. Stat items
        are holded until corresponding data item is received and vice versa.
        """
        raise NotImplementedError("Abstract method needs to be overridden")


class AbstractInfoWidget(object):
    ''' InfoWidgets interface
    parent class for all InfoWidgets'''

    def __init__(self):
        pass

    def render(self, screen):
        raise NotImplementedError("Abstract method needs to be overridden")

    def get_index(self):
        ''' get vertical priority index '''
        return 0


class AbstractCriterion(object):
    """ Criterions interface,
    parent class for all criterions """
    RC_TIME = 21
    RC_HTTP = 22
    RC_NET = 23
    RC_STEADY = 33

    def __init__(self):
        self.cause_second = None

    @staticmethod
    def count_matched_codes(codes_regex, codes_dict):
        """ helper to aggregate codes by mask """
        total = 0
        for code, count in codes_dict.items():
            if codes_regex.match(str(code)):
                total += count
        return total

    def notify(self, data, stat):
        """ notification about aggregate data goes here """
        raise NotImplementedError("Abstract methods requires overriding")

    def get_rc(self):
        """ get return code for test """
        raise NotImplementedError("Abstract methods requires overriding")

    def explain(self):
        """ long explanation to show after test stop """
        raise NotImplementedError("Abstract methods requires overriding")

    def widget_explain(self):
        """ short explanation to display in right panel """
        return self.explain(), 0

    @staticmethod
    def get_type_string():
        """ returns string that used as config name for criterion """
        raise NotImplementedError("Abstract methods requires overriding")


class GeneratorPlugin(object):
    DEFAULT_INFO = {
        'address': '',
        'port': 80,
        'instances': 1,
        'ammo_file': '',
        'rps_schedule': [],
        'duration': 0,
        'loop_count': 0
    }

    class Info(object):
        def __init__(
                self, address, port, instances, ammo_file, rps_schedule,
                duration, loop_count):
            self.address = address
            self.port = port
            self.instances = instances
            self.ammo_file = ammo_file
            self.rps_schedule = rps_schedule
            self.duration = duration
            self.loop_count = loop_count

    def get_info(self):
        # type: () -> Info
        return self.Info(**self.DEFAULT_INFO)



#! /usr/bin/env python2
""" The agent bundle, contains all metric classes and agent running code """
import ConfigParser
import commands
import base64
import logging
import os
import glob
import re
import socket
import subprocess
import sys
import time
import traceback
import signal
from threading import Thread
from optparse import OptionParser

logger = logging.getLogger(__name__)


def signal_handler(sig, frame):
    """ required for non-tty python runs to interrupt """
    logger.warning("Got signal %s", sig)
    raise KeyboardInterrupt()


def set_sig_handler():
    uncatchable = ['SIG_DFL', 'SIGSTOP', 'SIGKILL']
    for sig_name in [
            s for s in dir(signal)
            if (s.startswith("SIG") and s not in uncatchable)
    ]:
        try:
            sig_num = getattr(signal, sig_name)
            signal.signal(sig_num, signal_handler)
        except Exception as e:
            logger.error("Can't set handler for %s, %s", sig_name, e)


class AbstractMetric:
    """ Parent class for all metrics """

    def __init__(self):
        pass

    def columns(self):
        """ methods should return list of columns provided by metric class """
        raise NotImplementedError()

    def check(self):
        """ methods should return list of values provided by metric class """
        raise NotImplementedError()


class CpuLa(AbstractMetric):
    def columns(self, ):
        return ['System_la1', 'System_la5', 'System_la15']

    def check(self, ):
        loadavg_file = open('/proc/loadavg', 'r')
        loadavg_data = loadavg_file.readline().strip().split()[:3]
        result = map(str, loadavg_data)
        loadavg_file.close()
        return result


class CpuStat(AbstractMetric):
    """ read /proc/stat and calculate amount of time
        the CPU has spent performing different kinds of work.
    """

    def __init__(self):
        AbstractMetric.__init__(self)
        self.prev_check = {}
        self.current_check = {}

    def columns(self, ):
        columns = [
            'System_csw', 'System_int', 'CPU_user', 'CPU_nice', 'CPU_system',
            'CPU_idle', 'CPU_iowait', 'CPU_irq', 'CPU_softirq',
            'System_numproc', 'System_numthreads'
        ]
        return columns

    def check(self, ):
        # resulting data array
        result = []

        try:
            proc_stat_file = open('/proc/stat', 'r')
            proc_stat_all = proc_stat_file.readlines()
            proc_stat_file.close()
        except Exception as exc:
            logger.error(
                'Error opening /proc/stat. Traceback: %s',
                traceback.format_exc(exc))
            result.append([''] * 9)
        else:
            # Parse data
            try:
                for stat in proc_stat_all:
                    if stat.startswith('cpu '):
                        self.current_check['cpu'] = map(
                            float, stat.split("\n")[0].split()[1:8])
                    if stat.startswith('ctxt '):
                        self.current_check['csw'] = float(stat.split()[1])
                    if stat.startswith('intr '):
                        self.current_check['intr'] = float(stat.split()[1])
            except Exception as exc:
                logger.error(
                    'Error parsing /proc/stat data. Traceback: %s',
                    traceback.format_exc(exc))

            # Context switches and interrups delta
            try:
                if not self.prev_check.get('csw') or not self.prev_check.get(
                        'intr'):
                    self.prev_check['csw'] = self.current_check['csw']
                    self.prev_check['intr'] = self.current_check['intr']
                    result.extend([''] * 2)
                else:
                    delta_csw = str(
                        self.current_check['csw'] - self.prev_check['csw'])
                    delta_intr = str(
                        self.current_check['intr'] - self.prev_check['intr'])
                    self.prev_check['csw'] = self.current_check['csw']
                    self.prev_check['intr'] = self.current_check['intr']
                    result.append(delta_csw)
                    result.append(delta_intr)
            except Exception as exc:
                logger.error(
                    'Error trying to count delta of cpu interrupts and csw. Traceback: %s',
                    traceback.format_exc(exc))
                result.extend([''] * 2)

            # CPU usage metrics delta
            try:
                if not self.prev_check.get('cpu'):
                    self.prev_check['cpu'] = self.current_check['cpu']
                    result.extend([''] * 7)
                else:
                    delta = []
                    cnt = 0
                    sum_val = 0
                    for metric in self.current_check['cpu']:
                        delta_cpu = self.current_check['cpu'][
                            cnt] - self.prev_check['cpu'][cnt]
                        sum_val += delta_cpu
                        delta.append(delta_cpu)
                        cnt += 1
                    cnt = 0
                    for metric in self.current_check['cpu']:
                        result.append(str((delta[cnt] / sum_val) * 100))
                        cnt += 1
                    self.prev_check['cpu'] = self.current_check['cpu']
            except Exception as exc:
                logger.error(
                    'Error trying to count delta of cpu usage metrics. Traceback: %s',
                    traceback.format_exc(exc))
                result.extend([''] * 7)

        # Numproc
        try:
            proc_dirs = os.listdir('/proc/')
            pids = []
            for element in proc_dirs:
                try:
                    element = int(element)
                except Exception as exc:
                    pass
                else:
                    pids.append(element)
        except Exception as exc:
            logger.error(
                'Error trying to count numprocs. Traceback: %s',
                traceback.format_exc(exc))
            result.append([''])
        else:
            result.append(str(len(pids)))

        # Numthreads
        try:
            loadavg_file = open('/proc/loadavg', 'r')
            numthreads = loadavg_file.readline().split()[3].split('/')[1]
            loadavg_file.close()
        except Exception as exc:
            logger.error(
                'Error opening /proc/loadavg to get numthreads. Traceback: %s',
                traceback.format_exc(exc))
            result.append([''])
        else:
            result.append(numthreads)

        # Sample : ['localhost', '1437493895', '1088.0', '728.0', '0.0', '0.0',
        # '6.25390869293', '93.6835522201', '0.0', '0.0', '0.0625390869293',
        # '239', '534']
        return result


def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


class Custom(AbstractMetric):
    """ custom metrics: call and tail """

    def __init__(self, call, tail):
        AbstractMetric.__init__(self)
        self.call = call
        self.tail = tail
        self.diff_values = {}

    def columns(self, ):
        cols = []
        for el in self.tail:
            cols.append("Custom:" + el)
        for el in self.call:
            cols.append("Custom:" + el)
        return cols

    def check(self, ):
        res = []
        for el in self.tail:
            cmnd = base64.b64decode(el.split(':')[1])
            logger.debug("Run custom check: tail -n 1 %s", cmnd)
            output = subprocess.Popen(['tail', '-n', '1', cmnd],
                                      stdout=subprocess.PIPE).communicate()[0]
            res.append(self.diff_value(el, output.strip()))
        for el in self.call:
            cmnd = base64.b64decode(el.split(':')[1])

            logger.debug("Run custom check: %s", cmnd)
            output = subprocess.Popen(
                cmnd, shell=True, stdout=subprocess.PIPE).stdout.read()
            res.append(self.diff_value(el, output.strip()))
        logger.debug("Collected:\n%s", res)
        return res

    def diff_value(self, config_line, value):
        if not is_number(value):
            value = 0
        params = config_line.split(':')
        if len(params) > 2 and int(params[2]):
            if config_line in self.diff_values:
                diffvalue = float(value) - self.diff_values[config_line]
            else:
                diffvalue = 0
            self.diff_values[config_line] = float(value)
            value = diffvalue
        return str(value)


class Disk(AbstractMetric):
    def __init__(self):
        AbstractMetric.__init__(self)
        self.read = 0
        self.write = 0
        self.devs = self._get_devs()

    def columns(self, ):
        return ['Disk_read', 'Disk_write']

    def check(self, ):
        # cluster size
        size = 512

        # Current data
        read, writed = 0, 0

        try:
            with open("/proc/diskstats") as mfd:
                stat = mfd.readlines()

            for el in stat:
                data = el.split()
                if data[2] in self.devs:
                    read += int(data[5])
                    writed += int(data[9])

            if self.read or self.write:
                result = [
                    str(size * (read - self.read)),
                    str(size * (writed - self.write))
                ]
            else:
                result = ['', '']

            self.read, self.write = read, writed

        except Exception as exc:
            logger.error("%s: %s", exc, traceback.format_exc(exc))
            result = ['', '']
        return result

    @staticmethod
    def _get_devs():
        try:
            with open("/proc/mounts") as mfd:
                mounts = mfd.readlines()
        except IOError as exc:
            logger.exception("Can't read block devices")
            return []
        logger.info("Mounts: %s", mounts)
        devs = []
        for mount in mounts:
            if mount.startswith("/dev"):
                parts = mount.split(" ")
                rp = os.path.realpath(parts[0])
                short_name = rp.split(os.sep)[-1]
                devs.append(short_name)
                # Fixed due to LVM volumes on dom0 machines
                # here we're trying to get block device name (even if mounted
                # device file exists)
                try:
                    for dirc in glob.glob("/sys/devices/virtual/block/*"):
                        logger.debug("Checking %s", dirc)
                        name_path = "%s/dm/name" % dirc
                        if os.path.exists(name_path):
                            logger.debug("Checking %s", dirc)
                            try:
                                with open(name_path) as fds:
                                    nam = fds.read().strip()
                                    logger.info("Test: %s/%s", nam, short_name)
                                    if nam == short_name:
                                        dsk_name = dirc.split(os.sep)[-1]
                                        logger.info("Found: %s", dsk_name)
                                        devs.append(dsk_name)
                                        break
                            except Exception as exc:
                                logger.info(
                                    "Failed: %s", traceback.format_exc(exc))
                except Exception as exc:
                    logger.info(
                        "Failed to get block device name via /sys/devices/: %s",
                        traceback.format_exc(exc))
        logger.info("Devs: %s", devs)
        return devs


class Mem(AbstractMetric):
    """
    Memory statistics
    """
    empty = ''

    def __init__(self):
        AbstractMetric.__init__(self)
        self.name = 'advanced memory usage'
        self.vars = (
            'MemUsed', 'Buffers', 'Cached', 'MemFree', 'Dirty', 'MemTotal')

    #        self.open('/proc/meminfo')

    def columns(self):
        columns = [
            'Memory_total', 'Memory_used', 'Memory_free', 'Memory_shared',
            'Memory_buff', 'Memory_cached'
        ]
        logger.info("Start. Columns: %s" % columns)
        return columns

    def check(self):
        result = []
        try:
            with open('/proc/meminfo') as f:
                meminfo = f.readlines()
            data = {}
            for name in self.vars:
                data[name] = 0
            for l in meminfo:
                if ':' not in l:
                    continue
                [name, raw_value] = l.split(':')
                data.update({name: long(raw_value.split()[0]) / 1024.0})
            data['MemUsed'] = data['MemTotal'] - data['MemFree'] - data[
                'Buffers'] - data['Cached']
            result = [
                data['MemTotal'], data['MemUsed'], data['MemFree'], 0,
                data['Buffers'], data['Cached']
            ]
        except Exception as e:
            logger.error("Can't get meminfo, %s", e, exc_info=True)
            result.append([self.empty] * 9)
        return map(str, result)


class NetRetrans(AbstractMetric):
    """ read netstat output and
    calculate tcp segment retransmition derivative """

    def __init__(self):
        AbstractMetric.__init__(self)
        self.retr_second = None
        self.retr_first = None
        self.fetch = None
        self.delta = []

    def columns(self, ):
        return ['Net_retransmit', ]

    def check(self, ):
        self.fetch = lambda: int(commands.getoutput('netstat -s | grep "segments retransmited" | awk \'{print $1}\''))
        if self.retr_second is not None:
            self.retr_first = self.fetch()
            self.delta = []
            self.delta.append(str(self.retr_first - self.retr_second))
            self.retr_second = self.retr_first
            return self.delta
        else:
            # first check
            self.retr_second = self.fetch()
            return ['0', ]


class NetTcp(AbstractMetric):
    """ Read ss util output and count TCP socket's number grouped by state """

    def __init__(self):
        AbstractMetric.__init__(self)
        self.fields = [
            'Net_closewait',
            'Net_estab',
            'Net_timewait',
        ]
        self.keys = [
            'closed',
            'estab',
            'timewait',
        ]

    def columns(self, ):
        return self.fields

    def check(self, ):
        """
        * check is there TCP connections in "field" state in last check
        if note set it to 0.
        * make output ordered as "fields" list
        """

        def fetch():
            return commands.getoutput("ss -s | sed -ne '/^TCP:/p'")

        regex = ('(^[^(]+\()' '([^)]+)')
        matches = re.match(regex, fetch())
        raw = matches.group(2)

        data = {}
        result = []

        for i in raw.split(','):
            state, count = i.split()
            data[state] = count.split('/')[0]
        for field in self.keys:
            if field in data:
                result.append(str(data[field]))
            else:
                result.append('0')
        return result


class NetTxRx(AbstractMetric):
    """ Get upped iface names and read they Tx/Rx counters in bytes """

    def __init__(self):
        AbstractMetric.__init__(self)
        self.prev_rx = 0
        self.prev_tx = 0

    def columns(self, ):
        return [
            'Net_tx',
            'Net_rx',
        ]

    def check(self, ):
        """
        get network interface name which have ip addr
        which resolved fron  host FQDN.
        If we have network bonding or need to collect multiple iface
        statistic beter to change that behavior.
        """
        status, data = commands.getstatusoutput("/sbin/ifconfig -s")
        logger.debug("/sbin/ifconfig output is: %s", data)

        rx, tx = 0, 0

        if status == 0:
            try:
                lines = data.split('\n')

                def position(sample):
                    return lines[0].split().index(sample)

                rx_pos = position('RX-OK')
                tx_pos = position('TX-OK')

                for line in lines[1:]:
                    counters = line.split()
                    if counters[rx_pos].isdigit() and counters[tx_pos].isdigit(
                    ):
                        rx += int(counters[rx_pos])
                        tx += int(counters[tx_pos])
            except Exception as e:
                logger.error('Failed to parse ifconfig output %s: %s', data, e)

        logger.debug("Total RX/TX packets counters: %s", [str(rx), str(tx)])

        if self.prev_rx == 0:
            t_tx = 0
            t_rx = 0
        else:
            t_rx = rx - self.prev_rx
            t_tx = tx - self.prev_tx
        self.prev_rx = rx
        self.prev_tx = tx

        return [str(t_rx), str(t_tx)]


class Net(AbstractMetric):
    def __init__(self):
        AbstractMetric.__init__(self)
        self.recv = 0
        self.send = 0
        self.rgx = re.compile('\S+\s(\d+)\s(\d+)')

    def columns(self, ):
        return ['Net_recv', 'Net_send']

    def check(self, ):
        # Current data
        recv, send = 0, 0

        # TODO: change to simple file reading
        cmnd = "cat /proc/net/dev | tail -n +3 | cut -d':' -f 1,2 --output-delimiter=' ' | awk '{print $1, $2, $10}'"
        logger.debug("Starting: %s", cmnd)
        try:
            stat = subprocess.Popen([cmnd],
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE,
                                    shell=True)
        except Exception as exc:
            logger.error("Error getting net metrics: %s", exc)
            result = ['', '']

        else:
            err = stat.stderr.read()
            if err:
                logger.error("Error output: %s", err)
                result = ['', '']
            else:
                for elm in stat.stdout:
                    match = self.rgx.match(elm)
                    if match:
                        recv += int(match.group(1))
                        send += int(match.group(2))
                        logger.debug("Recv/send: %s/%s", recv, send)
                    else:
                        logger.debug("Not matched: %s", elm)
                if self.recv:
                    result = [str(recv - self.recv), str(send - self.send)]
                else:
                    result = ['', '']

        self.recv, self.send = recv, send
        logger.debug("Network recieved/sent bytes: %s", result)
        return result


# ===========================


class AgentWorker(Thread):
    dlmtr = ';'

    def __init__(self):
        Thread.__init__(self)
        self.last_run_ts = None
        self.startup_processes = []
        self.c_interval = 1
        self.tails = None
        self.calls = None
        self.metrics_collected = []
        self.startups = []
        self.shutdowns = []
        self.c_host = None
        self.c_local_start = None
        self.c_start = None
        self.daemon = True  # Thread auto-shutdown
        self.finished = False
        # metrics we know about
        self.known_metrics = {
            'cpu-la': CpuLa(),
            'cpu-stat': CpuStat(),
            'mem': Mem(),
            'net-retrans': NetRetrans(),
            'net-tx-rx': NetTxRx(),
            'net-tcp': NetTcp(),
            'disk': Disk(),
            'net': Net(),
        }

    @staticmethod
    def popen(cmnd):
        return subprocess.Popen(
            cmnd,
            bufsize=0,
            preexec_fn=os.setsid,
            close_fds=True,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)

    def run(self):
        logger.info("Running startup commands")
        for cmnd in self.startups:
            logger.debug("Run: %s", cmnd)
            proc = self.popen(cmnd)
            self.startup_processes.append(proc)

        logger.info("Start polling thread")
        header = []

        sync_time = str(self.c_start + (int(time.time()) - self.c_local_start))
        header.extend(
            ['start', self.c_host, sync_time])  # start compile init header

        # add metrics from config file to header
        for metric_name in self.metrics_collected:
            if metric_name:
                header.extend(self.known_metrics[metric_name].columns())

        # add custom metrics from config file to header
        custom = Custom(self.calls, self.tails)
        header.extend(custom.columns())

        sys.stdout.write(self.dlmtr.join(header) + '\n')
        sys.stdout.flush()

        logger.debug(self.dlmtr.join(header))

        # check loop
        while not self.finished:
            logger.debug('Start check')
            line = []
            sync_time = str(
                self.c_start + (int(time.time()) - self.c_local_start))
            line.extend([self.c_host, sync_time])

            # known metrics
            for metric_name in self.metrics_collected:
                if not metric_name:
                    continue
                try:
                    data = self.known_metrics[metric_name].check()
                    if len(data) != len(
                            self.known_metrics[metric_name].columns()):
                        raise RuntimeError(
                            "Data len not matched columns count: %s" % data)
                except Exception as e:
                    logger.exception('Can\'t fetch %s', metric_name)
                    data = ['0'
                            ] * len(self.known_metrics[metric_name].columns())
                line.extend(data)

            logger.debug("line: %s" % line)
            # custom commands
            line.extend(custom.check())

            # print result line
            try:
                row = self.dlmtr.join(line)
                logger.debug("str: %s" % row)
                sys.stdout.write(row + '\n')
                sys.stdout.flush()
            except IOError as e:
                logger.error("Can't send data to collector, terminating, %s", e)
                self.finished = True

            self.fixed_sleep(self.c_interval)

        logger.info("Terminating startup commands")
        for proc in self.startup_processes:
            logger.debug("Terminate: %s", proc)
            os.killpg(proc.pid, signal.SIGTERM)

        logger.info("Running shutdown commands")
        for cmnd in self.shutdowns:
            logger.debug("Run: %s", cmnd)
            subprocess.call(cmnd, shell=True)

        logger.info("Worker thread finished")

    def fixed_sleep(self, slp_interval):
        """ sleep 'interval' exclude processing time part """
        delay = slp_interval
        if self.last_run_ts is not None:
            delta = time.time() - self.last_run_ts
            delay = slp_interval - delta
            logger.debug("Sleep for: %s (delta %s)", delay, delta)

        time.sleep(delay if delay > 0 else 0)
        self.last_run_ts = time.time()


class AgentConfig:
    def __init__(self, def_cfg_path):
        self.c_interval = 1
        self.c_host = socket.getfqdn()
        logger.info("Start agent at host: %s\n" % self.c_host)
        self.c_start = None
        self.c_local_start = int(time.time())
        self.metrics_collected = []
        self.calls = []
        self.tails = []
        self.startups = []
        self.shutdowns = []

        options = self.parse_options(def_cfg_path)
        self.parse_config(options.cfg_file)

    def parse_options(self, def_cfg_path):
        # parse options
        parser = OptionParser()
        parser.add_option(
            '-c',
            '--config',
            dest='cfg_file',
            type='str',
            help='Config file path, default is: ./' + def_cfg_path,
            default=def_cfg_path)

        parser.add_option(
            '-t',
            '--timestamp',
            dest='timestamp',
            type='int',
            help='Caller timestamp for synchronization',
            default=self.c_local_start)
        (options, args) = parser.parse_args()

        self.c_start = options.timestamp
        logger.debug("Caller timestamp: %s", options.timestamp)

        return options

    def parse_config(self, cfg_file='agent.cfg'):
        # parse cfg file
        config = ConfigParser.ConfigParser()
        config.readfp(open(cfg_file))

        # metric section
        if config.has_option('metric', 'names'):
            self.metrics_collected = config.get('metric', 'names').split(',')

        # main section
        if config.has_section('main'):
            if config.has_option('main', 'interval'):
                self.c_interval = config.getfloat('main', 'interval')
            if config.has_option('main', 'host'):
                self.c_host = config.get('main', 'host')
            if config.has_option('main', 'start'):
                self.c_start = config.getint('main', 'start')

        logger.info('Agent params: %s, %s' % (self.c_interval, self.c_host))

        # custom section
        if config.has_section('custom'):
            if config.has_option('custom', 'tail'):
                self.tails += config.get('custom', 'tail').split(',')
            if config.has_option('custom', 'call'):
                self.calls += config.get('custom', 'call').split(',')

        if config.has_section('startup'):
            for option in config.options('startup'):
                if option.startswith('cmd'):
                    self.startups.append(config.get('startup', option))

        if config.has_section('shutdown'):
            for option in config.options('shutdown'):
                if option.startswith('cmd'):
                    self.shutdowns.append(config.get('shutdown', option))

    def prepare_worker(self, wrk):
        # populate
        wrk.c_start = self.c_start
        wrk.c_local_start = self.c_local_start
        wrk.c_interval = self.c_interval
        wrk.c_host = self.c_host
        wrk.metrics_collected = self.metrics_collected
        wrk.calls = self.calls
        wrk.tails = self.tails
        wrk.startups = self.startups
        wrk.shutdowns = self.shutdowns


if __name__ == '__main__':
    fname = os.path.dirname(__file__) + "_agent.log"
    level = logging.DEBUG

    fmt = "%(asctime)s - %(filename)s - %(name)s - %(levelname)s - %(message)s"
    logging.basicConfig(filename=fname, level=level, format=fmt)

    set_sig_handler()

    worker = AgentWorker()

    agent_config = AgentConfig('agent.cfg')
    agent_config.prepare_worker(worker)

    worker.start()

    try:
        logger.debug("Check for stdin shutdown command")
        cmd = sys.stdin.readline()
        if cmd:
            logger.info("Stdin cmd received: %s", cmd)
            worker.finished = True
    except KeyboardInterrupt:
        logger.info("Interrupted")
        worker.finished = True
    except:
        logger.error("Unexpected exception, try to shutdown", exc_info=True)
        worker.finished = True
    while worker.isAlive():
        try:
            logger.debug("Join the worker thread, waiting for cleanup")
            worker.join(10)
            if worker.isAlive():
                logger.error(
                    "Worker have not finished shutdown in "
                    "10 seconds, going to exit anyway")
                sys.exit(1)
        except KeyboardInterrupt:
            if not worker.isAlive():
                logger.info("Worker finished")
            else:
                logger.info("Waiting for worker process was interrupted")



"""
Split incoming DataFrames into chunks, cache them, union chunks with same key
and pass to the underlying aggregator.
"""

import pandas as pd


class TimeChopper(object):
    """
    TimeChopper splits incoming dataframes by index. Chunks are cached and
    chunks for same key from different DFs are joined. Then chunks are passed
    further as (<timestamp>, <dataframe>) tuples.
    """

    def __init__(self, source, cache_size):
        self.cache_size = cache_size
        self.source = source
        self.cache = {}

    def __iter__(self):
        for chunk in self.source:
            grouped = chunk.groupby(level=0)
            for group_key, group_data in list(grouped):
                if group_key in self.cache:
                    self.cache[group_key] = pd.concat(
                        [self.cache[group_key], group_data])
                else:
                    self.cache[group_key] = group_data
                while len(self.cache) > self.cache_size:
                    key = min(self.cache.keys())
                    yield (key, self.cache.pop(key, None))
        while self.cache:
            key = min(self.cache.keys())
            yield (key, self.cache.pop(key, None))



class MavenReader(object):
    def close(self):
        pass

    def __iter__(self):
        yield None


class MavenStatsReader(object):
    def close(self):
        pass

    def __iter__(self):
        yield None



"""Module to provide target monitoring"""

import datetime
import fnmatch
import logging
import os
import time

from pkg_resources import resource_string
from ...common.util import expand_to_seconds
from ...common.interfaces import AbstractPlugin,\
    MonitoringDataListener, AbstractInfoWidget, AbstractCriterion

from .collector import MonitoringCollector, MonitoringDataDecoder
from ..Autostop import Plugin as AutostopPlugin
from ..Console import Plugin as ConsolePlugin
from ..Phantom import Plugin as PhantomPlugin

logger = logging.getLogger(__name__)


class Plugin(AbstractPlugin):
    """  resource mon plugin  """

    SECTION = 'monitoring'

    def __init__(self, core):
        AbstractPlugin.__init__(self, core)
        self.jobno = None
        self.default_target = None
        self.config = None
        self.process = None
        self.monitoring = MonitoringCollector()
        self.die_on_fail = True
        self.data_file = None
        self.mon_saver = None
        self.address_resolver = None

    @staticmethod
    def get_key():
        return __file__

    def start_test(self):
        if self.monitoring:
            self.monitoring.load_start_time = time.time()
            logger.debug(
                "load_start_time = %s" % self.monitoring.load_start_time)

    def get_available_options(self):
        return ["config", "default_target", 'ssh_timeout']

    def configure(self):
        self.config = self.get_option("config", 'auto').strip()
        self.default_target = self.get_option("default_target", 'localhost')
        self.monitoring.ssh_timeout = expand_to_seconds(
            self.get_option('ssh_timeout', "5s"))

        if self.config == 'none' or self.config == 'auto':
            self.die_on_fail = False
        else:
            if self.config and self.config[0] == '<':
                xmlfile = self.core.mkstemp(".xml", "monitoring_")
                self.core.add_artifact_file(xmlfile)
                xml = open(xmlfile, 'w')
                xml.write(self.config)
                xml.close()
                self.config = xmlfile

            if not os.path.exists(self.config):
                raise OSError(
                    "Monitoring config file not found: %s" % self.config)

        if self.config == 'none':
            self.monitoring = None

        if self.config == 'auto':
            default_config = resource_string(
                __name__, 'config/monitoring_default_config.xml')
            self.config = self.core.mkstemp(".xml", "monitoring_default_")
            with open(self.config, 'w') as cfg_file:
                cfg_file.write(default_config)

        try:
            autostop = self.core.get_plugin_of_type(AutostopPlugin)
            autostop.add_criterion_class(MetricHigherCriterion)
            autostop.add_criterion_class(MetricLowerCriterion)
        except KeyError:
            logger.debug(
                "No autostop plugin found, not adding instances criterion")

    def prepare_test(self):
        try:
            phantom = self.core.get_plugin_of_type(PhantomPlugin)
            if phantom.phout_import_mode:
                logger.info("Phout import mode, disabling monitoring")
                self.config = None
                self.monitoring = None

            info = phantom.get_info()
            if info:
                self.default_target = info.address
                logger.debug(
                    "Changed monitoring target to %s", self.default_target)
        except KeyError as ex:
            logger.debug("Phantom plugin not found: %s", ex)

        if self.address_resolver:
            try:
                self.default_target = self.address_resolver.resolve_virtual(
                    self.default_target)
            except Exception as exc:
                logger.error("Failed to get target info: %s", exc)

        if not self.config or self.config == 'none':
            logger.info("Monitoring has been disabled")
        else:
            logger.info("Starting monitoring with config: %s", self.config)
            self.core.add_artifact_file(self.config, True)
            self.monitoring.config = self.config
            if self.default_target:
                self.monitoring.default_target = self.default_target

            self.data_file = self.core.mkstemp('.data', 'monitoring_')
            self.mon_saver = SaveMonToFile(self.data_file)
            self.monitoring.add_listener(self.mon_saver)
            self.core.add_artifact_file(self.data_file)

            try:
                console = self.core.get_plugin_of_type(ConsolePlugin)
            except Exception as ex:
                logger.debug("Console not found: %s", ex)
                console = None
            if console:
                widget = MonitoringWidget(self)
                console.add_info_widget(widget)
                self.monitoring.add_listener(widget)

            try:
                self.monitoring.prepare()
                self.monitoring.start()
                count = 0
                while not self.monitoring.first_data_received and count < 15 * 5:
                    time.sleep(0.2)
                    self.monitoring.poll()
                    count += 1
            except:
                logger.error("Could not start monitoring", exc_info=True)
                if self.die_on_fail:
                    raise
                else:
                    self.monitoring = None

    def is_test_finished(self):
        if self.monitoring:
            data_len = self.monitoring.poll()
            logger.debug("Monitoring got %s lines", data_len)
        return -1

    def end_test(self, retcode):
        logger.info("Finishing monitoring")
        if self.monitoring:
            self.monitoring.stop()
            for log in self.monitoring.artifact_files:
                self.core.add_artifact_file(log)

            while self.monitoring.send_data:
                logger.info("Sending monitoring data rests...")
                self.monitoring.send_collected_data()
        if self.mon_saver:
            self.mon_saver.close()
        return retcode


class SaveMonToFile(MonitoringDataListener):
    """
    Default listener - saves data to file
    """

    def __init__(self, out_file):
        MonitoringDataListener.__init__(self)
        if out_file:
            self.store = open(out_file, 'w')

    def monitoring_data(self, data_list):
        [self.store.write(data) for data in data_list]
        self.store.flush()

    def close(self):
        """ close open files """
        logger.debug("Closing monitoring file")
        if self.store:
            self.store.close()


class MonitoringWidget(
        AbstractInfoWidget, MonitoringDataListener, MonitoringDataDecoder):
    """
    Screen widget
    """

    def __init__(self, owner):
        AbstractInfoWidget.__init__(self)
        MonitoringDataDecoder.__init__(self)
        self.owner = owner
        self.data = {}
        self.sign = {}
        self.time = {}
        self.max_metric_len = 0

    def get_index(self):
        return 50

    def __handle_data_item(self, host, data):
        """ store metric in data tree and calc offset signs """
        for metric, value in data.iteritems():
            if value == '' or value == self.NA:
                value = self.NA
                self.sign[host][metric] = -1
                self.data[host][metric] = value
            else:
                if self.data[host][metric] == self.NA:
                    self.sign[host][metric] = 1
                elif float(value) > float(self.data[host][metric]):
                    self.sign[host][metric] = 1
                elif float(value) < float(self.data[host][metric]):
                    self.sign[host][metric] = -1
                else:
                    self.sign[host][metric] = 0
                self.data[host][metric] = "%.2f" % float(value)

    def monitoring_data(self, data_list):
        logger.debug("Mon widget data: %s", data_list)
        for line in data_list:
            if not line.strip():
                continue

            host, data, initial, timestamp = self.decode_line(line)
            self.time[host] = timestamp
            if initial:
                self.sign[host] = {}
                self.data[host] = {}
                for metric in data.keys():
                    self.sign[host][metric] = 0
                    self.data[host][metric] = self.NA
            else:
                self.__handle_data_item(host, data)

    def render(self, screen):
        if not self.owner.monitoring:
            return "Monitoring is " + screen.markup.RED + "offline" + screen.markup.RESET
        else:
            res = "Monitoring is " + screen.markup.GREEN + \
                "online" + screen.markup.RESET + ":\n"
            for hostname, metrics in self.data.items():
                tm_stamp = datetime.datetime.fromtimestamp(
                    float(self.time[hostname])).strftime('%H:%M:%S')
                res += (
                    "   " + screen.markup.CYAN + "%s" + screen.markup.RESET +
                    " at %s:\n") % (hostname, tm_stamp)
                for metric, value in sorted(metrics.iteritems()):
                    if self.sign[hostname][metric] > 0:
                        value = screen.markup.YELLOW + value + screen.markup.RESET
                    elif self.sign[hostname][metric] < 0:
                        value = screen.markup.CYAN + value + screen.markup.RESET
                    res += "      %s%s: %s\n" % (
                        ' ' * (self.max_metric_len - len(metric)),
                        metric.replace('_', ' '), value)

            return res.strip()


class AbstractMetricCriterion(
        AbstractCriterion, MonitoringDataListener, MonitoringDataDecoder):
    """ Parent class for metric criterion """

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        MonitoringDataDecoder.__init__(self)

        try:
            # FIXME telegraf update required here
            self.mon = autostop.core.get_plugin_of_type(Plugin)
            if self.mon.monitoring:
                self.mon.monitoring.add_listener(self)
        except KeyError:
            logger.warning("No monitoring module, mon autostop disabled")
        self.triggered = False
        self.autostop = autostop

        self.host = param_str.split(',')[0].strip()
        self.metric = param_str.split(',')[1].strip()
        self.value_limit = float(param_str.split(',')[2])
        self.seconds_limit = expand_to_seconds(param_str.split(',')[3])
        self.last_second = None
        self.seconds_count = 0

    def monitoring_data(self, data_list):
        if self.triggered:
            return

        for line in data_list:
            if not line.strip():
                continue

            host, data, initial, timestamp = self.decode_line(line)
            if initial or not fnmatch.fnmatch(host, self.host):
                continue

            if self.metric not in data.keys() or not data[self.metric] or data[
                    self.metric] == self.NA:
                data[self.metric] = 0

            logger.debug(
                "Compare %s %s/%s=%s to %s",
                self.get_type_string(), host, self.metric, data[self.metric],
                self.value_limit)
            if self.comparison_fn(float(data[self.metric]), self.value_limit):
                if not self.seconds_count:
                    self.cause_second = self.last_second

                logger.debug(self.explain())

                self.seconds_count += 1
                if self.seconds_count >= self.seconds_limit:
                    logger.debug("Triggering autostop")
                    self.triggered = True
                    return
            else:
                self.seconds_count = 0

    def notify(self, data, stat):
        if self.seconds_count:
            self.autostop.add_counting(self)

        self.last_second = (data, stat)
        return self.triggered

    def comparison_fn(self, arg1, arg2):
        """ comparison function """
        raise NotImplementedError()


class MetricHigherCriterion(AbstractMetricCriterion):
    """ trigger if metric is higher than limit """

    def __init__(self, autostop, param_str):
        AbstractMetricCriterion.__init__(self, autostop, param_str)

    def get_rc(self):
        return 31

    @staticmethod
    def get_type_string():
        return 'metric_higher'

    def explain(self):
        items = (self.host, self.metric, self.value_limit, self.seconds_count)
        return "%s/%s metric value is higher than %s for %s seconds" % items

    def widget_explain(self):
        items = (
            self.host, self.metric, self.value_limit, self.seconds_count,
            self.seconds_limit)
        return "%s/%s > %s for %s/%ss" % items, float(
            self.seconds_count) / self.seconds_limit

    def comparison_fn(self, arg1, arg2):
        return arg1 > arg2


class MetricLowerCriterion(AbstractMetricCriterion):
    """ trigger if metric is lower than limit """

    def __init__(self, autostop, param_str):
        AbstractMetricCriterion.__init__(self, autostop, param_str)

    def get_rc(self):
        return 32

    @staticmethod
    def get_type_string():
        return 'metric_lower'

    def explain(self):
        items = (self.host, self.metric, self.value_limit, self.seconds_count)
        return "%s/%s metric value is lower than %s for %s seconds" % items

    def widget_explain(self):
        items = (
            self.host, self.metric, self.value_limit, self.seconds_count,
            self.seconds_limit)
        return "%s/%s < %s for %s/%ss" % items, float(
            self.seconds_count) / self.seconds_limit

    def comparison_fn(self, arg1, arg2):
        return arg1 < arg2


class AbstractResolver:
    """ Resolver class provides virtual to real host resolution """

    def __init__(self):
        pass

    def resolve_virtual(self, virt_address):
        """ get host address by virtual """
        raise NotImplementedError()



import logging
import subprocess
import time

from ...common.interfaces import AbstractPlugin

logger = logging.getLogger(__name__)


class Plugin(AbstractPlugin):
    ''' Start appium before the test, stop after it ended '''

    SECTION = "appium"

    def __init__(self, core):
        super(Plugin, self).__init__(core)
        self.appium_cmd = None
        self.appium_log = None
        self.appium_port = None
        self.process_stdout = None

    @staticmethod
    def get_key():
        return __file__

    def get_available_options(self):
        opts = ["appium_cmd"]
        return opts

    def configure(self):
        # plugin part
        self.appium_cmd = self.get_option("appium_cmd", "appium")
        self.appium_user = self.get_option("user", "")
        self.appium_port = self.get_option("port", "4723")
        self.appium_log = self.core.mkstemp(".log", "appium_")
        self.core.add_artifact_file(self.appium_log)

    def prepare_test(self):
        args = [self.appium_cmd, '-p', self.appium_port, '-g', self.appium_log]
        if self.appium_user:
            args = ["su", "-c", " ".join(args), self.appium_user]
        logger.info("Starting appium server: %s", args)
        self.process_start_time = time.time()
        process_stdout_file = self.core.mkstemp(".log", "appium_stdout_")
        self.core.add_artifact_file(process_stdout_file)
        self.process_stdout = open(process_stdout_file, 'w')
        self.process = subprocess.Popen(
            args,
            stderr=self.process_stdout,
            stdout=self.process_stdout,
            close_fds=True)
        logger.info("Waiting 5 seconds for Appium to start...")
        time.sleep(5)

    def is_test_finished(self):
        retcode = self.process.poll()
        if retcode is not None:
            logger.warning("Appium exited: %s", retcode)
            return abs(retcode)
        else:
            return -1

    def end_test(self, retcode):
        if self.process and self.process.poll() is None:
            logger.info(
                "Terminating appium process with PID %s", self.process.pid)
            self.process.terminate()
            if self.process_stdout:
                self.process_stdout.close()
        else:
            logger.warn("Appium finished unexpectedly")
        return retcode

    def get_info(self):
        return None



# TODO: make the next two lines unnecessary
# pylint: disable=line-too-long
# pylint: disable=missing-docstring
import json
import logging
import os

from ...common.interfaces import AbstractPlugin,\
    MonitoringDataListener, AggregateResultListener

logger = logging.getLogger(__name__)  # pylint: disable=C0103


class Plugin(AbstractPlugin, AggregateResultListener, MonitoringDataListener):
    # pylint:disable=R0902
    SECTION = 'json_report'

    def __init__(self, core):
        super(Plugin, self).__init__(core)
        self._is_telegraf = None

    def post_process(self, retcode):
        try:
            if self.aggregator_data_logger:
                for handle in self.aggregator_data_logger.handlers:
                    handle.close()
            if self.monitoring_logger:
                for handle in self.monitoring_logger.handlers:
                    handle.close()
        except AttributeError:
            logger.error("Logger wasn't initialized yet", exc_info=True)
        return retcode

    def get_available_options(self):
        return ['monitoring_log', 'test_data_log', 'test_stats_log']

    def configure(self):
        self.monitoring_logger = self.create_file_logger(
            'monitoring', self.get_option('monitoring_log', 'monitoring.log'))
        self.aggregator_data_logger = self.create_file_logger(
            'aggregator_data',
            self.get_option('test_data_log', 'test_data.log'))
        self.core.job.subscribe_plugin(self)

    def create_file_logger(self, logger_name, file_name, formatter=None):
        loggr = logging.getLogger(logger_name)
        loggr.setLevel(logging.INFO)
        handler = logging.FileHandler(
            os.path.join(self.core.artifacts_dir, file_name), mode='w')
        handler.setLevel(logging.INFO)
        if formatter:
            handler.setFormatter(formatter)
        loggr.addHandler(handler)
        loggr.propagate = False
        return loggr

    def on_aggregated_data(self, data, stats):
        """
        @data: aggregated data
        @stats: stats about gun
        """
        self.aggregator_data_logger.info(
            json.dumps({
                'data': data,
                'stats': stats
            }))

    def monitoring_data(self, data_list):
        if self.is_telegraf:
            self.monitoring_logger.info(json.dumps(data_list))
        else:
            [
                self.monitoring_logger.info(data.strip()) for data in data_list
                if data
            ]

    @property
    def is_telegraf(self):
        if self._is_telegraf is None:
            self._is_telegraf = 'Telegraf' in self.core.job.monitoring_plugin.__module__
        return self._is_telegraf



""" The central part of the tool: Core """
import datetime
import fnmatch
import importlib as il
import json
import logging
import os
import shutil
import socket
import tempfile
import time
import traceback
import uuid
import pkg_resources
import sys
import platform
from builtins import str

from yandextank.common.exceptions import PluginNotPrepared
from yandextank.common.interfaces import GeneratorPlugin

from ..common.util import update_status, execute, pid_exists

from ..common.resource import manager as resource
from ..plugins.Aggregator import Plugin as AggregatorPlugin
from ..plugins.Monitoring import Plugin as MonitoringPlugin
from ..plugins.Telegraf import Plugin as TelegrafPlugin
if sys.version_info[0] < 3:
    import ConfigParser
else:
    import configparser as ConfigParser

logger = logging.getLogger(__name__)


class Job(object):
    def __init__(
            self,
            name,
            description,
            task,
            version,
            config_copy,
            monitoring_plugin,
            aggregator_plugin,
            tank,
            generator_plugin=None):
        # type: (unicode, unicode, unicode, unicode, unicode, MonitoringPlugin,
        # AggregatorPlugin, GeneratorPlugin) -> Job
        self.name = name
        self.description = description
        self.task = task
        self.version = version
        self.config_copy = config_copy
        self.monitoring_plugin = monitoring_plugin
        self.aggregator_plugin = aggregator_plugin
        self.tank = tank
        self._phantom_info = None
        self.generator_plugin = generator_plugin

    def subscribe_plugin(self, plugin):
        self.aggregator_plugin.add_result_listener(plugin)
        try:
            self.monitoring_plugin.monitoring.add_listener(plugin)
        except AttributeError:
            logging.warning('Monitoring plugin is not enabled')

    @property
    def phantom_info(self):
        if self._phantom_info is None:
            raise PluginNotPrepared
        return self._phantom_info

    @phantom_info.setter
    def phantom_info(self, info):
        self._phantom_info = info


class TankCore(object):
    """
    JMeter + dstat inspired :)
    """
    SECTION = 'tank'
    SECTION_META = 'meta'
    PLUGIN_PREFIX = 'plugin_'
    PID_OPTION = 'pid'
    UUID_OPTION = 'uuid'
    LOCK_DIR = '/var/lock'

    def __init__(self, artifacts_base_dir=None, artifacts_dir_name=None):
        self.config = ConfigManager()
        self.status = {}
        self.plugins = []
        self.artifacts_dir_name = artifacts_dir_name
        self._artifacts_dir = None
        self.artifact_files = {}
        self.artifacts_base_dir = artifacts_base_dir
        self.manual_start = False
        self.scheduled_start = None
        self.interrupted = False
        self.lock_file = None
        self.flush_config_to = None
        self.lock_dir = None
        self.taskset_path = None
        self.taskset_affinity = None
        self.uuid = str(uuid.uuid4())
        self.set_option(self.SECTION, self.UUID_OPTION, self.uuid)
        self.set_option(self.SECTION, self.PID_OPTION, str(os.getpid()))
        self.job = None

    def get_uuid(self):
        return self.uuid

    def get_available_options(self):
        return [
            "artifacts_base_dir", "artifacts_dir", "flush_config_to",
            "taskset_path", "affinity"
        ]

    def load_configs(self, configs):
        """ Tells core to load configs set into options storage """
        logger.info("Loading configs...")
        self.config.load_files(configs)
        dotted_options = []
        for option, value in self.config.get_options(self.SECTION):
            if '.' in option:
                dotted_options += [option + '=' + value]
        self.apply_shorthand_options(dotted_options, self.SECTION)
        self.config.flush()
        self.add_artifact_file(self.config.file)
        self.set_option(self.SECTION, self.PID_OPTION, str(os.getpid()))
        self.flush_config_to = self.get_option(
            self.SECTION, "flush_config_to", "")
        if self.flush_config_to:
            self.config.flush(self.flush_config_to)

    def load_plugins(self):
        """
        Tells core to take plugin options and instantiate plugin classes
        """
        logger.info("Loading plugins...")

        if not self.artifacts_base_dir:
            self.artifacts_base_dir = os.path.expanduser(
                self.get_option(self.SECTION, "artifacts_base_dir", '.'))

        if not self.artifacts_dir_name:
            self.artifacts_dir_name = self.get_option(
                self.SECTION, "artifacts_dir", "")

        self.taskset_path = self.get_option(
            self.SECTION, 'taskset_path', 'taskset')
        self.taskset_affinity = self.get_option(self.SECTION, 'affinity', '')

        options = self.config.get_options(self.SECTION, self.PLUGIN_PREFIX)
        for (plugin_name, plugin_path) in options:
            if not plugin_path:
                logger.debug("Seems the plugin '%s' was disabled", plugin_name)
                continue
            logger.debug("Loading plugin %s from %s", plugin_name, plugin_path)
            # FIXME cleanup an old deprecated plugin path format
            if '/' in plugin_path:
                logger.warning(
                    "Deprecated plugin path format: %s\n"
                    "Should be in pythonic format. Example:\n"
                    "    plugin_jmeter=yandextank.plugins.JMeter", plugin_path)
                if plugin_path.startswith("Tank/Plugins/"):
                    plugin_path = "yandextank.plugins." + \
                                  plugin_path.split('/')[-1].split('.')[0]
                    logger.warning("Converted plugin path to %s", plugin_path)
                else:
                    raise ValueError(
                        "Couldn't convert plugin path to new format:\n    %s" %
                        plugin_path)
            try:
                plugin = il.import_module(plugin_path)
            except ImportError:
                if plugin_path.startswith("yatank_internal_"):
                    logger.warning(
                        "Deprecated plugin path format: %s\n"
                        "Tank plugins are now orginized using"
                        " namespace packages. Example:\n"
                        "    plugin_jmeter=yandextank.plugins.JMeter",
                        plugin_path)
                    plugin_path = plugin_path.replace(
                        "yatank_internal_", "yandextank.plugins.")
                if plugin_path.startswith("yatank_"):
                    logger.warning(
                        "Deprecated plugin path format: %s\n"
                        "Tank plugins are now orginized using"
                        " namespace packages. Example:\n"
                        "    plugin_jmeter=yandextank.plugins.JMeter",
                        plugin_path)

                    plugin_path = plugin_path.replace(
                        "yatank_", "yandextank.plugins.")
                logger.warning("Patched plugin path: %s", plugin_path)
                plugin = il.import_module(plugin_path)
            try:
                instance = getattr(plugin, 'Plugin')(self)
            except:
                logger.warning(
                    "Deprecated plugin classname: %s. Should be 'Plugin'",
                    plugin)
                instance = getattr(
                    plugin, plugin_path.split('.')[-1] + 'Plugin')(self)

            self.plugins.append(instance)

        logger.debug("Plugin instances: %s", self.plugins)

    def plugins_configure(self):
        """        Call configure() on all plugins        """
        self.publish("core", "stage", "configure")
        if not os.path.exists(self.artifacts_base_dir):
            os.makedirs(self.artifacts_base_dir)

            os.chmod(self.artifacts_base_dir, 0o755)

        logger.info("Configuring plugins...")
        if self.taskset_affinity != '':
            self.taskset(os.getpid(), self.taskset_path, self.taskset_affinity)

        # monitoring plugin
        try:
            mon = self.get_plugin_of_type(TelegrafPlugin)
        except KeyError:
            logger.debug("Telegraf plugin not found:", exc_info=True)
            try:
                mon = self.get_plugin_of_type(MonitoringPlugin)
            except KeyError:
                logger.debug("Monitoring plugin not found:", exc_info=True)
                mon = None

        # aggregator plugin
        try:
            aggregator = self.get_plugin_of_type(AggregatorPlugin)
        except KeyError:
            logger.warning("Aggregator plugin not found:", exc_info=True)
            aggregator = None

        # generator plugin
        try:
            gen = self.get_plugin_of_type(GeneratorPlugin)
        except KeyError:
            logger.warning("Load generator not found:", exc_info=True)
            gen = None

        self.job = Job(
            name=self.get_option(self.SECTION_META, "job_name",
                                 'none').decode('utf8'),
            description=self.get_option(self.SECTION_META, "job_dsc",
                                        '').decode('utf8'),
            task=self.get_option(self.SECTION_META, 'task',
                                 'dir').decode('utf8'),
            version=self.get_option(self.SECTION_META, 'ver',
                                    '').decode('utf8'),
            config_copy=self.get_option(
                self.SECTION_META, 'copy_config_to', 'config_copy'),
            monitoring_plugin=mon,
            aggregator_plugin=aggregator,
            generator_plugin=gen,
            tank=socket.getfqdn())

        for plugin in self.plugins:
            logger.debug("Configuring %s", plugin)
            plugin.configure()
            self.config.flush()
        if self.flush_config_to:
            self.config.flush(self.flush_config_to)

    def plugins_prepare_test(self):
        """ Call prepare_test() on all plugins        """
        logger.info("Preparing test...")
        self.publish("core", "stage", "prepare")
        for plugin in self.plugins:
            logger.debug("Preparing %s", plugin)
            plugin.prepare_test()
        if self.flush_config_to:
            self.config.flush(self.flush_config_to)

    def plugins_start_test(self):
        """        Call start_test() on all plugins        """
        logger.info("Starting test...")
        self.publish("core", "stage", "start")
        for plugin in self.plugins:
            logger.debug("Starting %s", plugin)
            plugin.start_test()
        if self.flush_config_to:
            self.config.flush(self.flush_config_to)

    def wait_for_finish(self):
        """
        Call is_test_finished() on all plugins 'till one of them initiates exit
        """

        logger.info("Waiting for test to finish...")
        self.publish("core", "stage", "shoot")
        if not self.plugins:
            raise RuntimeError("It's strange: we have no plugins loaded...")

        while not self.interrupted:
            begin_time = time.time()
            for plugin in self.plugins:
                logger.debug("Polling %s", plugin)
                retcode = plugin.is_test_finished()
                if retcode >= 0:
                    return retcode
            end_time = time.time()
            diff = end_time - begin_time
            logger.debug("Polling took %s", diff)
            logger.debug("Tank status:\n%s", json.dumps(self.status, indent=2))
            # screen refresh every 0.5 s
            if diff < 0.5:
                time.sleep(0.5 - diff)
        return 1

    def plugins_end_test(self, retcode):
        """        Call end_test() on all plugins        """
        logger.info("Finishing test...")
        self.publish("core", "stage", "end")

        for plugin in self.plugins:
            logger.debug("Finalize %s", plugin)
            try:
                logger.debug("RC before: %s", retcode)
                plugin.end_test(retcode)
                logger.debug("RC after: %s", retcode)
            except Exception as ex:
                logger.error("Failed finishing plugin %s: %s", plugin, ex)
                logger.debug(
                    "Failed finishing plugin: %s", traceback.format_exc(ex))
                if not retcode:
                    retcode = 1

        if self.flush_config_to:
            self.config.flush(self.flush_config_to)
        return retcode

    def plugins_post_process(self, retcode):
        """
        Call post_process() on all plugins
        """
        logger.info("Post-processing test...")
        self.publish("core", "stage", "post_process")

        for plugin in self.plugins:
            logger.debug("Post-process %s", plugin)
            try:
                logger.debug("RC before: %s", retcode)
                retcode = plugin.post_process(retcode)
                logger.debug("RC after: %s", retcode)
            except Exception as ex:
                logger.error("Failed post-processing plugin %s: %s", plugin, ex)
                logger.debug(
                    "Failed post-processing plugin: %s",
                    traceback.format_exc(ex))
                if not retcode:
                    retcode = 1

        if self.flush_config_to:
            self.config.flush(self.flush_config_to)

        self.__collect_artifacts()

        return retcode

    def taskset(self, pid, path, affinity):
        if affinity:
            args = "%s -pc %s %s" % (path, affinity, pid)
            retcode, stdout, stderr = execute(
                args, shell=True, poll_period=0.1, catch_out=True)
            logger.debug('taskset stdout: %s', stdout)
            if retcode != 0:
                raise KeyError(stderr)
            else:
                logger.info(
                    "Enabled taskset for pid %s with affinity %s",
                    str(pid), affinity)

    def __collect_artifacts(self):
        logger.debug("Collecting artifacts")
        logger.info("Artifacts dir: %s", self.artifacts_dir)
        for filename, keep in self.artifact_files.items():
            try:
                self.__collect_file(filename, keep)
            except Exception as ex:
                logger.warn("Failed to collect file %s: %s", filename, ex)

    def get_option(self, section, option, default=None):
        """
        `Get` an option from option storage
        and `set` if default specified.
        """
        if not self.config.config.has_section(section):
            logger.debug("No section '%s', adding", section)
            self.config.config.add_section(section)

        try:
            value = self.config.config.get(section, option).strip()
        except ConfigParser.NoOptionError as ex:
            if default is not None:
                default = str(default)
                self.config.config.set(section, option, default)
                self.config.flush()
                value = default.strip()
            else:
                logger.warn(
                    "Mandatory option %s was not found in section %s", option,
                    section)
                raise ex

        if len(value) > 1 and value[0] == '`' and value[-1] == '`':
            logger.debug("Expanding shell option %s", value)
            retcode, stdout, stderr = execute(value[1:-1], True, 0.1, True)
            if retcode or stderr:
                raise ValueError(
                    "Error expanding option %s, RC: %s" % (value, retcode))
            value = stdout.strip()

        return value

    def set_option(self, section, option, value):
        """
        Set an option in storage
        """
        if not self.config.config.has_section(section):
            self.config.config.add_section(section)
        self.config.config.set(section, option, value)
        self.config.flush()

    def get_plugin_of_type(self, plugin_class):
        """
        Retrieve a plugin of desired class, KeyError raised otherwise
        """
        logger.debug("Searching for plugin: %s", plugin_class)
        matches = [
            plugin for plugin in self.plugins
            if isinstance(plugin, plugin_class)
        ]
        if len(matches) > 0:
            if len(matches) > 1:
                logger.debug(
                    "More then one plugin of type %s found. Using first one.",
                    plugin_class)
            return matches[-1]
        else:
            raise KeyError("Requested plugin type not found: %s" % plugin_class)

    def __collect_file(self, filename, keep_original=False):
        """
        Move or copy single file to artifacts dir
        """
        dest = self.artifacts_dir + '/' + os.path.basename(filename)
        logger.debug("Collecting file: %s to %s", filename, dest)
        if not filename or not os.path.exists(filename):
            logger.warning("File not found to collect: %s", filename)
            return

        if os.path.exists(dest):
            # FIXME: 3 find a way to store artifacts anyway
            logger.warning("File already exists: %s", dest)
            return

        if keep_original:
            shutil.copy(filename, self.artifacts_dir)
        else:
            shutil.move(filename, self.artifacts_dir)

        os.chmod(dest, 0o644)

    def add_artifact_file(self, filename, keep_original=False):
        """
        Add file to be stored as result artifact on post-process phase
        """
        if filename:
            logger.debug(
                "Adding artifact file to collect (keep=%s): %s", keep_original,
                filename)
            self.artifact_files[filename] = keep_original

    def apply_shorthand_options(self, options, default_section='DEFAULT'):
        for option_str in options:
            try:
                section = option_str[:option_str.index('.')]
                option = option_str[option_str.index('.') + 1:option_str.index(
                    '=')]
            except ValueError:
                section = default_section
                option = option_str[:option_str.index('=')]
            value = option_str[option_str.index('=') + 1:]
            logger.debug(
                "Override option: %s => [%s] %s=%s", option_str, section,
                option, value)
            self.set_option(section, option, value)

    def get_lock_dir(self):
        if not self.lock_dir:
            self.lock_dir = self.get_option(
                self.SECTION, "lock_dir", self.LOCK_DIR)

        return os.path.expanduser(self.lock_dir)

    def get_lock(self, force=False):
        if not force and self.__there_is_locks():
            raise RuntimeError("There is lock files")

        fh, self.lock_file = tempfile.mkstemp(
            '.lock', 'lunapark_', self.get_lock_dir())
        os.close(fh)
        os.chmod(self.lock_file, 0o644)
        self.config.file = self.lock_file
        self.config.flush()

    def release_lock(self):
        self.config.file = None
        if self.lock_file and os.path.exists(self.lock_file):
            logger.debug("Releasing lock: %s", self.lock_file)
            os.remove(self.lock_file)

    def __there_is_locks(self):
        retcode = False
        lock_dir = self.get_lock_dir()
        for filename in os.listdir(lock_dir):
            if fnmatch.fnmatch(filename, 'lunapark_*.lock'):
                full_name = os.path.join(lock_dir, filename)
                logger.warn("Lock file present: %s", full_name)

                try:
                    info = ConfigParser.ConfigParser()
                    info.read(full_name)
                    pid = info.get(TankCore.SECTION, self.PID_OPTION)
                    if not pid_exists(int(pid)):
                        logger.debug(
                            "Lock PID %s not exists, ignoring and "
                            "trying to remove", pid)
                        try:
                            os.remove(full_name)
                        except Exception as exc:
                            logger.debug(
                                "Failed to delete lock %s: %s", full_name, exc)
                    else:
                        retcode = True
                except Exception as exc:
                    logger.warn(
                        "Failed to load info from lock %s: %s", full_name, exc)
                    retcode = True
        return retcode

    def mkstemp(self, suffix, prefix, directory=None):
        """
        Generate temp file name in artifacts base dir
        and close temp file handle
        """
        if not directory:
            directory = self.artifacts_base_dir
        fd, fname = tempfile.mkstemp(suffix, prefix, directory)
        os.close(fd)
        os.chmod(fname, 0o644)  # FIXME: chmod to parent dir's mode?
        return fname

    def publish(self, publisher, key, value):
        update_status(self.status, [publisher] + key.split('.'), value)

    def close(self):
        """
        Call close() for all plugins
        """
        logger.info("Close allocated resources...")

        for plugin in self.plugins:
            logger.debug("Close %s", plugin)
            try:
                plugin.close()
            except Exception as ex:
                logger.error("Failed closing plugin %s: %s", plugin, ex)
                logger.debug(
                    "Failed closing plugin: %s", traceback.format_exc(ex))

    @property
    def artifacts_dir(self):
        if not self._artifacts_dir:
            if not self.artifacts_dir_name:
                date_str = datetime.datetime.now().strftime(
                    "%Y-%m-%d_%H-%M-%S.")
                self.artifacts_dir_name = tempfile.mkdtemp(
                    "", date_str, self.artifacts_base_dir)
            elif not os.path.isdir(self.artifacts_dir_name):
                os.makedirs(self.artifacts_dir_name)
            os.chmod(self.artifacts_dir_name, 0o755)
            self._artifacts_dir = os.path.abspath(self.artifacts_dir_name)
        return self._artifacts_dir

    @staticmethod
    def get_user_agent():
        tank_agent = 'YandexTank/{}'.format(
            pkg_resources.require('yandextank')[0].version)
        py_info = sys.version_info
        python_agent = 'Python/{}.{}.{}'.format(
            py_info[0], py_info[1], py_info[2])
        os_agent = 'OS/{}'.format(platform.platform())
        return ' '.join((tank_agent, python_agent, os_agent))


class ConfigManager(object):
    """ Option storage class """

    def __init__(self):
        self.file = None
        self.config = ConfigParser.ConfigParser()

    def load_files(self, configs):
        """         Read configs set into storage        """
        logger.debug("Reading configs: %s", configs)
        config_filenames = [
            resource.resource_filename(config) for config in configs
        ]
        try:
            self.config.read(config_filenames)
        except Exception as ex:
            logger.error("Can't load configs: %s", ex)
            raise ex

    def flush(self, filename=None):
        """        Flush current stat to file        """
        if not filename:
            filename = self.file

        if filename:
            with open(filename, 'w') as handle:
                self.config.write(handle)

    def get_options(self, section, prefix=''):
        """ Get options list with requested prefix """
        res = []
        try:
            for option in self.config.options(section):
                if not prefix or option.find(prefix) == 0:
                    res += [(
                        option[len(prefix):], self.config.get(section, option))]
        except ConfigParser.NoSectionError as ex:
            logger.warning("No section: %s", ex)

        logger.debug(
            "Section: [%s] prefix: '%s' options:\n%s", section, prefix, res)
        return res

    def find_sections(self, prefix):
        """ return sections with specified prefix """
        res = []
        for section in self.config.sections():
            if section.startswith(prefix):
                res.append(section)
        return res



''' Cummulative Autostops '''
import logging
import math
import re
from collections import deque

import numpy as np
from ...common.util import expand_to_milliseconds, expand_to_seconds

from .criterions import AbstractCriterion

logger = logging.getLogger(__name__)


class WindowCounter(object):
    def __init__(self, window_size):
        self.window_size = window_size
        self.value = 0.0
        self.q = deque()

    def push(self, value):
        self.value += value
        self.q.append(value)
        if len(self.q) > self.window_size:
            self.value -= self.q.popleft()

    def __len__(self):
        return len(self.q)


class TotalFracTimeCriterion(AbstractCriterion):
    '''
    Windowed time criterion

    syntax: total_time(300ms, 70%, 3s)

    Stop test if 70 percent of response times are greater then 300ms in
    a sliding window of 3 seconds.

    Warning: this criterion uses histogram to make its decision, so the
    time boundary is rounded to aggregator bin edge.
    '''

    @staticmethod
    def get_type_string():
        return 'total_time'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.autostop = autostop
        param = param_str.split(',')
        self.rt_limit = expand_to_milliseconds(param[0]) * 1000
        self.fail_ratio_limit = float(param[1][:-1]) / 100.0
        self.window_size = expand_to_seconds(param[2])
        self.fail_counter = WindowCounter(self.window_size)
        self.total_counter = WindowCounter(self.window_size)
        self.total_fail_ratio = 0.0
        self.seconds = deque()

    def __fail_count(self, data):
        ecdf = np.cumsum(data["overall"]["interval_real"]["hist"]["data"])
        idx = np.searchsorted(
            data["overall"]["interval_real"]["hist"]["bins"], self.rt_limit)
        if idx == 0:
            return ecdf[-1]
        elif idx == len(ecdf):
            return 0
        else:
            return ecdf[-1] - ecdf[idx]

    def notify(self, data, stat):
        self.seconds.append((data, stat))
        self.fail_counter.push(self.__fail_count(data))
        self.total_counter.push(data["overall"]["interval_real"]["len"])
        self.total_fail_ratio = (
            self.fail_counter.value / self.total_counter.value)
        if self.total_fail_ratio >= self.fail_ratio_limit and len(
                self.fail_counter) >= self.window_size:
            self.cause_second = self.seconds[0]
            logger.debug(self.explain())
            return True
        if len(self.seconds) > self.window_size:
            self.seconds.popleft()
        return False

    def get_rc(self):
        return 25

    def explain(self):
        return (
            "%.2f%% responses times higher "
            "than %sms for %ss since: %s" % (
                self.total_fail_ratio * 100, self.rt_limit / 1000,
                self.window_size, self.cause_second[0]["ts"]))

    def widget_explain(self):
        return (
            "%.2f%% times >%sms for %ss" % (
                self.total_fail_ratio * 100, self.rt_limit / 1000,
                self.window_size), self.total_fail_ratio)


class TotalHTTPCodesCriterion(AbstractCriterion):
    ''' Cummulative HTTP Criterion '''

    @staticmethod
    def get_type_string():
        return 'total_http'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        self.codes_mask = param_str.split(',')[0].lower()
        self.codes_regex = re.compile(self.codes_mask.replace("x", '.'))
        self.autostop = autostop
        self.data = deque()
        self.second_window = deque()

        level_str = param_str.split(',')[1].strip()
        if level_str[-1:] == '%':
            self.level = float(level_str[:-1])
            self.is_relative = True
        else:
            self.level = int(level_str)
            self.is_relative = False
        self.seconds_limit = expand_to_seconds(param_str.split(',')[2])

    def notify(self, data, stat):
        matched_responses = self.count_matched_codes(
            self.codes_regex, data["overall"]["proto_code"]["count"])
        if self.is_relative:
            if data["overall"]["interval_real"]["len"] > 0:
                matched_responses = float(matched_responses) / data["overall"][
                    "interval_real"]["len"] * 100
            else:
                matched_responses = 1
        logger.debug(
            "HTTP codes matching mask %s: %s/%s", self.codes_mask,
            matched_responses, self.level)
        self.data.append(matched_responses)
        self.second_window.append((data, stat))
        if len(self.data) > self.seconds_limit:
            self.data.popleft()
            self.second_window.popleft()
        queue_len = 1
        if self.is_relative:
            queue_len = len(self.data)
        if (sum(self.data) / queue_len) >= self.level\
                and len(self.data) >= self.seconds_limit:  # yapf:disable
            self.cause_second = self.second_window[0]
            logger.debug(self.explain())
            return True
        return False

    def get_rc(self):
        return 26

    def get_level_str(self):
        ''' format level str '''
        if self.is_relative:
            level_str = str(self.level) + "%"
        else:
            level_str = self.level
        return level_str

    def explain(self):
        if self.is_relative:
            items = (
                self.codes_mask, self.get_level_str(), self.seconds_limit,
                self.cause_second[0]["ts"])
            return (
                "%s codes count higher "
                "than %s for %ss, ended at: %s" % items)
        items = (
            self.codes_mask, self.get_level_str(), self.seconds_limit,
            self.cause_second[0]["ts"])
        return "%s codes count higher than %s for %ss, since %s" % items

    def widget_explain(self):
        if self.is_relative:
            items = (self.codes_mask, self.get_level_str(), self.seconds_limit)
            return ("HTTP %s>%s for %ss" % items, sum(self.data))
        items = (self.codes_mask, self.get_level_str(), self.seconds_limit)
        return ("HTTP %s>%s for %ss" % items, 1.0)


class TotalNetCodesCriterion(AbstractCriterion):
    ''' Cummulative Net Criterion '''

    @staticmethod
    def get_type_string():
        return 'total_net'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        self.codes_mask = param_str.split(',')[0].lower()
        self.codes_regex = re.compile(self.codes_mask.replace("x", '.'))
        self.autostop = autostop
        self.data = deque()
        self.second_window = deque()

        level_str = param_str.split(',')[1].strip()
        if level_str[-1:] == '%':
            self.level = float(level_str[:-1])
            self.is_relative = True
        else:
            self.level = int(level_str)
            self.is_relative = False
        self.seconds_limit = expand_to_seconds(param_str.split(',')[2])

    def notify(self, data, stat):
        codes = data["overall"]["net_code"]["count"].copy()
        if '0' in codes.keys():
            codes.pop('0')
        matched_responses = self.count_matched_codes(self.codes_regex, codes)
        if self.is_relative:
            if data["overall"]["interval_real"]["len"]:
                matched_responses = float(matched_responses) / data["overall"][
                    "interval_real"]["len"] * 100
                logger.debug(
                    "Net codes matching mask %s: %s%%/%s", self.codes_mask,
                    round(matched_responses, 2), self.get_level_str())
            else:
                matched_responses = 1
        else:
            logger.debug(
                "Net codes matching mask %s: %s/%s", self.codes_mask,
                matched_responses, self.get_level_str())

        self.data.append(matched_responses)
        self.second_window.append((data, stat))
        if len(self.data) > self.seconds_limit:
            self.data.popleft()
            self.second_window.popleft()

        queue_len = 1
        if self.is_relative:
            queue_len = len(self.data)
        if (sum(self.data) / queue_len) >= self.level\
                and len(self.data) >= self.seconds_limit:  # yapf:disable
            self.cause_second = self.second_window[0]
            logger.debug(self.explain())
            return True
        return False

    def get_rc(self):
        return 27

    def get_level_str(self):
        ''' format level str '''
        if self.is_relative:
            level_str = str(self.level) + "%"
        else:
            level_str = str(self.level)
        return level_str

    def explain(self):
        if self.is_relative:
            items = (
                self.codes_mask, self.get_level_str(), self.seconds_limit,
                self.cause_second[0]["ts"])
            return (
                "%s net codes count higher "
                "than %s for %ss, since %s" % items)
        items = (
            self.codes_mask, self.get_level_str(), self.seconds_limit,
            self.cause_second[0]["ts"])
        return "%s net codes count higher than %s for %ss, since %s" % items

    def widget_explain(self):
        if self.is_relative:
            items = (self.codes_mask, self.get_level_str(), self.seconds_limit)
            return ("Net %s>%s for %ss" % items, self.level)
        items = (self.codes_mask, self.get_level_str(), self.seconds_limit)
        return ("Net %s>%s for %ss" % items, self.level)


class TotalNegativeHTTPCodesCriterion(AbstractCriterion):
    ''' Reversed HTTP Criterion '''

    @staticmethod
    def get_type_string():
        return 'negative_http'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        self.codes_mask = param_str.split(',')[0].lower()
        self.codes_regex = re.compile(self.codes_mask.replace("x", '.'))
        self.autostop = autostop
        self.data = deque()
        self.second_window = deque()

        level_str = param_str.split(',')[1].strip()
        if level_str[-1:] == '%':
            self.level = float(level_str[:-1])
            self.is_relative = True
        else:
            self.level = int(level_str)
            self.is_relative = False
        self.seconds_limit = expand_to_seconds(param_str.split(',')[2])

    def notify(self, data, stat):
        matched_responses = self.count_matched_codes(
            self.codes_regex, data["overall"]["proto_code"]["count"])
        if self.is_relative:
            if data["overall"]["interval_real"]["len"]:
                matched_responses = float(matched_responses) / data["overall"][
                    "interval_real"]["len"] * 100
                matched_responses = 100 - matched_responses
            else:
                matched_responses = 1
            logger.debug(
                "HTTP codes matching mask not %s: %s/%s", self.codes_mask,
                round(matched_responses, 1), self.level)
        else:
            matched_responses = (
                data["overall"]["interval_real"]["len"] - matched_responses)
            logger.debug(
                "HTTP codes matching mask not %s: %s/%s", self.codes_mask,
                matched_responses, self.level)
        self.data.append(matched_responses)
        self.second_window.append((data, stat))
        if len(self.data) > self.seconds_limit:
            self.data.popleft()
            self.second_window.popleft()

        queue_len = 1
        if self.is_relative:
            queue_len = len(self.data)
        if (sum(self.data) / queue_len) >= self.level\
                and len(self.data) >= self.seconds_limit:  # yapf:disable
            self.cause_second = self.second_window[0]
            logger.debug(self.explain())
            return True
        return False

    def get_rc(self):
        return 28

    def get_level_str(self):
        ''' format level str'''
        if self.is_relative:
            level_str = str(self.level) + "%"
        else:
            level_str = self.level
        return level_str

    def explain(self):
        if self.is_relative:
            items = (
                self.codes_mask, self.get_level_str(), self.seconds_limit,
                self.cause_second[0]["ts"])
            return (
                "Not %s codes count higher "
                "than %s for %ss, since %s" % items)
        items = (
            self.codes_mask, self.get_level_str(), self.seconds_limit,
            self.cause_second[0]["ts"])
        return "Not %s codes count higher than %s for %ss, since %s" % items

    def widget_explain(self):
        if self.is_relative:
            items = (self.codes_mask, self.get_level_str(), self.seconds_limit)
            return ("HTTP not %s>%s for %ss" % items, sum(self.data))
        items = (self.codes_mask, self.get_level_str(), self.seconds_limit)
        return ("HTTP not %s>%s for %ss" % items, 1.0)


class TotalNegativeNetCodesCriterion(AbstractCriterion):
    ''' Reversed NET Criterion '''

    @staticmethod
    def get_type_string():
        return 'negative_net'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        self.codes_mask = param_str.split(',')[0].lower()
        self.codes_regex = re.compile(self.codes_mask.replace("x", '.'))
        self.autostop = autostop
        self.data = deque()
        self.second_window = deque()

        level_str = param_str.split(',')[1].strip()
        if level_str[-1:] == '%':
            self.level = float(level_str[:-1])
            self.is_relative = True
        else:
            self.level = int(level_str)
            self.is_relative = False
        self.seconds_limit = expand_to_seconds(param_str.split(',')[2])

    def notify(self, data, stat):
        codes = data["overall"]["net_code"]["count"].copy()
        # if '0' in codes.keys():
        #     codes.pop('0')
        matched_responses = self.count_matched_codes(self.codes_regex, codes)
        if self.is_relative:
            if data["overall"]["interval_real"]["len"]:
                matched_responses = float(matched_responses) / data["overall"][
                    "interval_real"]["len"] * 100
                matched_responses = 100 - matched_responses
            else:
                matched_responses = 1
            logger.debug(
                "Net codes matching mask not %s: %s/%s", self.codes_mask,
                round(matched_responses, 1), self.level)
        else:
            matched_responses = (
                data["overall"]["interval_real"]["len"] - matched_responses)
            logger.debug(
                "Net codes matching mask not %s: %s/%s", self.codes_mask,
                matched_responses, self.level)
        self.data.append(matched_responses)
        self.second_window.append((data, stat))
        if len(self.data) > self.seconds_limit:
            self.data.popleft()
            self.second_window.popleft()

        queue_len = 1
        if self.is_relative:
            queue_len = len(self.data)
        if (sum(self.data) / queue_len) >= self.level \
                and len(self.data) >= self.seconds_limit:  # yapf:disable
            self.cause_second = self.second_window[0]
            logger.debug(self.explain())
            return True
        return False

    def get_rc(self):
        return 29

    def get_level_str(self):
        ''' format level str'''
        if self.is_relative:
            level_str = str(self.level) + "%"
        else:
            level_str = self.level
        return level_str

    def explain(self):
        if self.is_relative:
            items = (
                self.codes_mask, self.get_level_str(), self.seconds_limit,
                self.cause_second[0]["ts"])
            return (
                "Not %s codes count higher "
                "than %s for %ss, since %s" % items)
        items = (
            self.codes_mask, self.get_level_str(), self.seconds_limit,
            self.cause_second[0]["ts"])
        return "Not %s codes count higher than %s for %ss, since %s" % items

    def widget_explain(self):
        if self.is_relative:
            items = (self.codes_mask, self.get_level_str(), self.seconds_limit)
            return ("Net not %s>%s for %ss" % items, sum(self.data))
        items = (self.codes_mask, self.get_level_str(), self.seconds_limit)
        return ("Net not %s>%s for %ss" % items, 1.0)


class TotalHTTPTrendCriterion(AbstractCriterion):
    ''' HTTP Trend Criterion '''

    @staticmethod
    def get_type_string():
        return 'http_trend'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        self.codes_mask = param_str.split(',')[0].lower()
        self.codes_regex = re.compile(self.codes_mask.replace("x", '.'))
        self.autostop = autostop
        self.tangents = deque()
        self.second_window = deque()
        self.total_tan = float()

        self.tangents.append(0)
        self.last = 0
        self.seconds_limit = expand_to_seconds(param_str.split(',')[1])
        self.measurement_error = float()

    def notify(self, data, stat):
        matched_responses = self.count_matched_codes(
            self.codes_regex, data["overall"]["proto_code"]["count"])

        self.tangents.append(matched_responses - self.last)
        self.second_window.append((data, stat))

        self.last = matched_responses

        if len(self.tangents) > self.seconds_limit:
            self.tangents.popleft()
            self.second_window.popleft()

        self.measurement_error = self.calc_measurement_error(self.tangents)

        self.total_tan = float(sum(self.tangents) / len(self.tangents))
        logger.debug(
            "Last trend for http codes %s: %.2f +/- %.2f", self.codes_mask,
            self.total_tan, self.measurement_error)

        if self.total_tan + self.measurement_error < 0:
            self.cause_second = self.second_window[0]
            logger.debug(self.explain())
            return True

        return False

    def calc_measurement_error(self, tangents):
        '''
        formula for measurement error
        sqrt ( (sum(1, n, (k_i - <k>)**2) / (n*(n-1)))
        '''

        if len(tangents) < 2:
            return 0.0

        avg_tan = float(sum(tangents) / len(tangents))
        numerator = float()
        for i in tangents:
            numerator += (i - avg_tan) * (i - avg_tan)

        return math.sqrt(numerator / len(tangents) / (len(tangents) - 1))

    def get_rc(self):
        return 30

    def explain(self):
        items = (
            self.codes_mask, self.total_tan, self.measurement_error,
            self.seconds_limit, self.cause_second[0]["ts"])
        return (
            "Last trend for %s http codes "
            "is %.2f +/- %.2f for %ss, since %s" % items)

    def widget_explain(self):
        items = (
            self.codes_mask, self.total_tan, self.measurement_error,
            self.seconds_limit)
        return ("HTTP(%s) trend is %.2f +/- %.2f < 0 for %ss" % items, 1.0)


class QuantileOfSaturationCriterion(AbstractCriterion):
    ''' Quantile of Saturation Criterion
        example: qsat(50ms, 3m, 10%) '''

    @staticmethod
    def get_type_string():
        return 'qsat'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        raise NotImplementedError

    #     self.autostop = autostop
    #     self.data = deque()
    #     self.second_window = deque()

    #     params = param_str.split(',')
    #     # qunatile in ms
    #     self.timing = expand_to_milliseconds(params[0])
    #     # width of time in seconds
    #     self.width = expand_to_seconds(params[1])
    #     # max height of deviations in percents
    #     self.height = float(params[2].split('%')[0])
    #     # last deviation in percents
    #     self.deviation = float()

    # def __get_timing_quantile(self, data):
    #     ''' get quantile level for criterion timing '''
    #     quan = 0.0
    #     for timing in sorted(aggr_data.cumulative.times_dist.keys()):
    #         timing_item = aggr_data.cumulative.times_dist[timing]
    #         quan += float(timing_item[
    #             'count']) / aggr_data.cumulative.total_count
    #         logger.debug("tt: %s %s", self.timing, timing_item['to'])
    #         if self.timing <= timing_item['to']:
    #             return quan
    #     return quan

    # def notify(self, data, stat):
    #     quan = 100 * self.__get_timing_quantile(data)
    #     logger.debug("Quantile for %s: %s", self.timing, quan)

    #     self.data.append(quan)
    #     self.second_window.append((data, stat))

    #     if len(self.data) > self.width:
    #         self.autostop.add_counting(self)
    #         self.data.popleft()
    #         self.second_window.popleft()

    #         self.deviation = max(self.data) - min(self.data)
    #         logger.debug(self.explain())

    #         if self.deviation < self.height:
    #             return True
    #     return False

    # def get_rc(self):
    #     return 33

    # def explain(self):
    #     items = (self.timing, self.width, self.deviation, self.height)
    #     return "%sms variance for %ss: %.3f%% (<%s%%)" % items

    # def widget_explain(self):
    #     level = self.height / self.deviation
    #     items = (self.timing, self.width, self.deviation, self.height)
    #     return ("%sms variance for %ss: %.3f%% (<%s%%)" % items, level)



import queue
import hashlib
import logging
import os
import subprocess
import tempfile
import threading
import time
from shutil import copyfile, rmtree

from ...common.util import SecuredShell

from ..Telegraf.config import AgentConfig
from ..Telegraf.reader import MonitoringReader

logger = logging.getLogger(__name__)
logging.getLogger("paramiko.transport").setLevel(logging.WARNING)


def generate_file_md5(filename, blocksize=2**20):
    m = hashlib.md5()
    with open(filename, "rb") as f:
        while True:
            buf = f.read(blocksize)
            if not buf:
                break
            m.update(buf)
    return m.hexdigest()


class LocalhostClient(object):
    """ localhost client setup """

    def __init__(self, config, old_style_configs):
        # config
        self.python = config['python']
        self.host = "localhost"
        self.telegraf = config['telegraf']
        self.config = AgentConfig(config, old_style_configs)

        # connection
        self.incoming_queue = queue.Queue()
        self.buffer = ""

        self.workdir = None
        self.reader = MonitoringReader(self.incoming_queue)

        self.path = {
            'AGENT_LOCAL_FOLDER': os.path.dirname(__file__) + '/agent',
            'TELEGRAF_LOCAL_PATH': self.telegraf,
        }

    def install(self):
        self.workdir = tempfile.mkdtemp()
        logger.info("Created temp dir %s", self.workdir)
        agent_config = self.config.create_collector_config(self.workdir)
        startup_config = self.config.create_startup_config()
        customs_script = self.config.create_custom_exec_script()
        try:
            copyfile(
                self.path['AGENT_LOCAL_FOLDER'] + '/agent.py',
                self.workdir + '/agent.py')
            copyfile(agent_config, self.workdir + '/agent.cfg')
            copyfile(startup_config, self.workdir + '/agent_startup.cfg')
            copyfile(customs_script, self.workdir + '/agent_customs.sh')
            if not os.path.isfile(self.path['TELEGRAF_LOCAL_PATH']):
                logger.error(
                    'Telegraf binary not found at specified path: %s\n'
                    'You can download telegraf binaries here: https://github.com/influxdata/telegraf\n'
                    'or install debian package: `telegraf`',
                    self.path['TELEGRAF_LOCAL_PATH'])
                return None, None, None
        except Exception:
            logger.error(
                "Failed to copy agent to %s on localhost",
                self.workdir,
                exc_info=True)
            return None, None, None
        return agent_config, startup_config, customs_script

    @staticmethod
    def popen(cmnd):
        return subprocess.Popen(
            cmnd,
            bufsize=0,
            preexec_fn=os.setsid,
            close_fds=True,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE, )

    def start(self):
        """Start local agent"""
        logger.info('Starting agent on localhost')
        command = "{python} {work_dir}/agent.py --telegraf {telegraf_path} --host {host}".format(
            python=self.python,
            work_dir=self.workdir,
            telegraf_path=self.path['TELEGRAF_LOCAL_PATH'],
            host=self.host)
        self.session = self.popen(command)
        self.reader_thread = threading.Thread(target=self.read_buffer)
        self.reader_thread.setDaemon(True)
        return self.session

    def read_buffer(self):
        while self.session:
            try:
                chunk = self.session.stdout.read(4096)
                if chunk:
                    parts = chunk.rsplit('\n', 1)
                    if len(parts) > 1:
                        ready_chunk = self.buffer + parts[0] + '\n'
                        self.buffer = parts[1]
                        self.incoming_queue.put(ready_chunk)
                    else:
                        self.buffer += parts[0]
            except ValueError:
                logger.debug(
                    'this exc most likely raised during interpreter shutdown\n'
                    'otherwise something really nasty happend',
                    exc_info=True)

    def uninstall(self):
        """
        Remove agent's files from remote host
        """
        if self.session:
            logger.info('Waiting monitoring data...')
            self.session.terminate()
            self.session.wait()
            self.session = None
        log_filename = "agent_{host}.log".format(host="localhost")
        data_filename = "agent_{host}.rawdata".format(host="localhost")
        try:
            logger.info('Saving monitoring artefacts from localhost')
            copyfile(self.workdir + "/_agent.log", log_filename)
            copyfile(self.workdir + "/monitoring.rawdata", data_filename)
            logger.info('Deleting temp directory: %s', self.workdir)
            rmtree(self.workdir)
        except Exception:
            logger.error("Exception while uninstalling agent", exc_info=True)

        logger.info("Removing agent from: localhost")
        return log_filename, data_filename


class SSHClient(object):
    """remote agent client setup """

    def __init__(self, config, old_style_configs, timeout):
        # config
        self.host = config['host']
        self.username = config['username']
        self.python = config['python']
        self.port = config['port']
        self.telegraf = config['telegraf']
        self.config = AgentConfig(config, old_style_configs)

        # connection
        self.session = None
        self.ssh = SecuredShell(self.host, self.port, self.username, timeout)
        self.incoming_queue = queue.Queue()
        self.buffer = ""

        self.reader = MonitoringReader(self.incoming_queue)
        handle, cfg_path = tempfile.mkstemp('.cfg', 'agent_')
        os.close(handle)
        self.path = {
            # Destination path on remote host
            'AGENT_REMOTE_FOLDER': '/tmp/',
            # Source path on tank
            'AGENT_LOCAL_FOLDER': os.path.dirname(__file__) + '/agent',
            'TELEGRAF_REMOTE_PATH': '/tmp/telegraf',
            'TELEGRAF_LOCAL_PATH': self.telegraf,
        }

    def install(self):
        """Create folder and copy agent and metrics scripts to remote host"""
        logger.info(
            "Installing monitoring agent at %s@%s...", self.username, self.host)

        # create remote temp dir
        cmd = self.python + ' -c "import tempfile; print tempfile.mkdtemp();"'
        logger.info("Creating temp dir on %s", self.host)
        try:
            out, errors, err_code = self.ssh.execute(cmd)
        except Exception:
            logger.error(
                "Failed to install monitoring agent to %s",
                self.host,
                exc_info=True)
            return None, None, None
        if errors:
            logging.error("[%s] error: '%s'", self.host, errors)
            return None, None, None

        if err_code:
            logging.error(
                "Failed to create remote dir via SSH at %s@%s, code %s: %s" %
                (self.username, self.host, err_code, out.strip()))
            return None, None, None

        remote_dir = out.strip()
        if remote_dir:
            self.path['AGENT_REMOTE_FOLDER'] = remote_dir
        logger.debug(
            "Remote dir at %s:%s", self.host, self.path['AGENT_REMOTE_FOLDER'])

        # create collector config
        agent_config = self.config.create_collector_config(
            self.path['AGENT_REMOTE_FOLDER'])
        startup_config = self.config.create_startup_config()
        customs_script = self.config.create_custom_exec_script()

        # trying to detect os version/architecture and get information about telegraf client
        # DO NOT DELETE indices in string format below. Python 2.6 does not
        # support string formatting without indices
        remote_cmd = 'import os; print os.path.isfile("' + self.path[
            'TELEGRAF_REMOTE_PATH'] + '")'
        cmd = self.python + ' -c \'{cmd}\''.format(cmd=remote_cmd)
        remote_telegraf_exists = "False"
        try:
            out, err, err_code = self.ssh.execute(cmd)
        except Exception:
            logger.error(
                "SSH execute error trying to check telegraf availability on host %s",
                self.host,
                exc_info=True)
        else:
            if err:
                logging.error("[%s] error: '%s'", self.host, errors)
            if out.strip():
                remote_telegraf_exists = out.strip()

        try:
            if remote_telegraf_exists in "True":
                logger.debug('Found telegraf client on %s..', self.host)
            else:
                logger.debug(
                    'Not found telegraf client on %s, trying to install from tank. Copying..',
                    self.host)
                if os.path.isfile(self.path['TELEGRAF_LOCAL_PATH']):
                    self.ssh.send_file(
                        self.path['TELEGRAF_LOCAL_PATH'],
                        self.path['TELEGRAF_REMOTE_PATH'])
                elif os.path.isfile("/usr/bin/telegraf"):
                    self.ssh.send_file(
                        '/usr/bin/telegraf', self.path['TELEGRAF_REMOTE_PATH'])
                else:
                    logger.error(
                        'Telegraf binary not found neither on %s nor on localhost at specified path: %s\n'
                        'You can download telegraf binaries here: https://github.com/influxdata/telegraf\n'
                        'or install debian package: `telegraf`', self.host,
                        self.path['TELEGRAF_LOCAL_PATH'])
                    return None, None, None

            self.ssh.send_file(
                self.path['AGENT_LOCAL_FOLDER'] + '/agent.py',
                self.path['AGENT_REMOTE_FOLDER'] + '/agent.py')
            self.ssh.send_file(
                agent_config, self.path['AGENT_REMOTE_FOLDER'] + '/agent.cfg')
            self.ssh.send_file(
                startup_config,
                self.path['AGENT_REMOTE_FOLDER'] + '/agent_startup.cfg')
            self.ssh.send_file(
                customs_script,
                self.path['AGENT_REMOTE_FOLDER'] + '/agent_customs.sh')

        except Exception:
            logger.error(
                "Failed to install agent on %s", self.host, exc_info=True)
            return None, None, None

        return agent_config, startup_config, customs_script

    def start(self):
        """Start remote agent"""
        logger.info('Starting agent: %s', self.host)
        command = "{python} {agent_path}/agent.py --telegraf {telegraf_path} --host {host}".format(
            python=self.python,
            agent_path=self.path['AGENT_REMOTE_FOLDER'],
            telegraf_path=self.path['TELEGRAF_REMOTE_PATH'],
            host=self.host)
        logging.debug('Command to start agent: %s', command)
        self.session = self.ssh.async_session(command)
        self.reader_thread = threading.Thread(target=self.read_buffer)
        self.reader_thread.setDaemon(True)
        return self.session

    def read_buffer(self):
        while self.session:
            chunk = self.session.read_maybe()
            if chunk:
                parts = chunk.rsplit('\n', 1)
                if len(parts) > 1:
                    ready_chunk = self.buffer + parts[0] + '\n'
                    self.buffer = parts[1]
                    self.incoming_queue.put(ready_chunk)
                else:
                    self.buffer += parts[0]
            else:
                time.sleep(1)

    def uninstall(self):
        """
        Remove agent's files from remote host
        """
        log_filename = "agent_{host}.log".format(host=self.host)
        data_filename = "agent_{host}.rawdata".format(host=self.host)

        try:
            if self.session:
                self.session.send("stop\n")
                self.session.close()
                self.session = None
        except:
            logger.warning(
                'Unable to correctly stop monitoring agent - session is broken. Pay attention to agent log (%s).',
                log_filename,
                exc_info=True)
        try:
            self.ssh.get_file(
                self.path['AGENT_REMOTE_FOLDER'] + "/_agent.log", log_filename)
            self.ssh.get_file(
                self.path['AGENT_REMOTE_FOLDER'] + "/monitoring.rawdata",
                data_filename)
            self.ssh.rm_r(self.path['AGENT_REMOTE_FOLDER'])
        except Exception:
            logger.error("Unable to get agent artefacts", exc_info=True)

        logger.info("Removing agent from: %s@%s...", self.username, self.host)
        return log_filename, data_filename



# -*- coding: utf-8 -*-
""" Provides class to run TankCore from python """
import logging
import os
import sys
import time
import traceback
import fnmatch
from pkg_resources import resource_filename

from ..core import tankcore


class ApiWorker:
    """    Worker class that runs tank core via python   """

    def __init__(self):
        self.core = tankcore.TankCore()
        self.baseconfigs_location = '/etc/yandex-tank'
        self.log = logging.getLogger(__name__)

    def init_logging(self, log_filename="tank.log"):
        """ Set up logging """
        logger = logging.getLogger('')
        self.log_filename = log_filename
        self.core.add_artifact_file(self.log_filename)

        file_handler = logging.FileHandler(self.log_filename)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(
            logging.Formatter(
                "%(asctime)s [%(levelname)s] %(name)s %(message)s"))
        logger.addHandler(file_handler)
        console_handler = logging.StreamHandler(sys.stdout)
        stderr_hdl = logging.StreamHandler(sys.stderr)

        # fmt_verbose = logging.Formatter(
        #     "%(asctime)s [%(levelname)s] %(name)s %(message)s")
        fmt_regular = logging.Formatter(
            "%(asctime)s %(levelname)s: %(message)s", "%H:%M:%S")

        console_handler.setLevel(logging.INFO)
        console_handler.setFormatter(fmt_regular)
        stderr_hdl.setFormatter(fmt_regular)

        f_err = SingleLevelFilter(logging.ERROR, True)
        f_warn = SingleLevelFilter(logging.WARNING, True)
        f_crit = SingleLevelFilter(logging.CRITICAL, True)
        console_handler.addFilter(f_err)
        console_handler.addFilter(f_warn)
        console_handler.addFilter(f_crit)
        logger.addHandler(console_handler)

        f_info = SingleLevelFilter(logging.INFO, True)
        f_debug = SingleLevelFilter(logging.DEBUG, True)
        stderr_hdl.addFilter(f_info)
        stderr_hdl.addFilter(f_debug)
        logger.addHandler(stderr_hdl)

    def __add_user_options(self):
        """ override config options with user specified options"""
        if self.options.get('user_options', None):
            self.core.apply_shorthand_options(self.options['user_options'])

    def configure(self, options):
        """ Make preparations before running Tank """
        self.options = options
        if self.options.get('lock_dir', None):
            self.core.set_option(
                self.core.SECTION, "lock_dir", self.options['lock_dir'])

        while True:
            try:
                self.core.get_lock(self.options.get('ignore_lock', None))
                break
            except Exception as exc:
                if self.options.get('lock_fail', None):
                    raise RuntimeError("Lock file present, cannot continue")
                self.log.info(
                    "Couldn't get lock. Will retry in 5 seconds... (%s)",
                    str(exc))
                time.sleep(5)

        configs = self.get_default_configs()
        if self.options.get('config', None):
            configs.append(self.options['config'])
        self.core.load_configs(configs)
        self.__add_user_options()
        self.core.load_plugins()

        if self.options.get('ignore_lock', None):
            self.core.set_option(self.core.SECTION, self.IGNORE_LOCKS, "1")

    def perform_test(self):
        """ Run the test and wait for finish """
        self.log.info("Performing test...")
        retcode = 1
        try:
            self.core.plugins_configure()
            self.core.plugins_prepare_test()
            if self.options.get('manual_start', None):
                self.log.info(
                    "Manual start option specified, waiting for user actions")
                raw_input("Press Enter key to start test")

            self.core.plugins_start_test()
            retcode = self.core.wait_for_finish()
            retcode = self.core.plugins_end_test(retcode)
            retcode = self.core.plugins_post_process(retcode)
        except KeyboardInterrupt as ex:
            self.log.info(
                "Do not press Ctrl+C again, the test will be broken otherwise")
            self.log.debug(
                "Caught KeyboardInterrupt: %s", traceback.format_exc(ex))
            try:
                retcode = self.__graceful_shutdown()
            except KeyboardInterrupt as ex:
                self.log.debug(
                    "Caught KeyboardInterrupt again: %s",
                    traceback.format_exc(ex))
                self.log.info(
                    "User insists on exiting, aborting graceful shutdown...")
                retcode = 1

        except Exception as ex:
            self.log.info("Exception: %s", traceback.format_exc(ex))
            self.log.error("%s", ex)
            retcode = self.__graceful_shutdown()
            self.core.release_lock()

        self.log.info("Done performing test with code %s", retcode)
        return retcode

    def get_default_configs(self):
        """ returns default configs list, from /etc, home dir and package_data"""
        # initialize basic defaults
        configs = [resource_filename(__name__, 'config/00-base.ini')]
        try:
            conf_files = sorted(os.listdir(self.baseconfigs_location))
            for filename in conf_files:
                if fnmatch.fnmatch(filename, '*.ini'):
                    configs += [
                        os.path.realpath(
                            self.baseconfigs_location + os.sep + filename)
                    ]
        except OSError:
            self.log.warn(
                self.baseconfigs_location +
                ' is not accessible to get configs list')

        configs += [os.path.expanduser('~/.yandex-tank')]
        return configs

    def __graceful_shutdown(self):
        """ call shutdown routines """
        retcode = 1
        self.log.info("Trying to shutdown gracefully...")
        retcode = self.core.plugins_end_test(retcode)
        retcode = self.core.plugins_post_process(retcode)
        self.log.info("Done graceful shutdown")
        return retcode


class SingleLevelFilter(logging.Filter):
    """Exclude or approve one msg type at a time.    """

    def __init__(self, passlevel, reject):
        logging.Filter.__init__(self)
        self.passlevel = passlevel
        self.reject = reject

    def filter(self, record):
        if self.reject:
            return record.levelno != self.passlevel
        else:
            return record.levelno == self.passlevel



""" Contains Phantom Plugin, Console widgets, result reader classes """
# FIXME: 3 there is no graceful way to interrupt the process of phout import
# TODO: phout import
import logging
import multiprocessing as mp
import os
import subprocess
import time

from ...common.util import execute, expand_to_seconds
from ...common.interfaces import AbstractPlugin, AbstractCriterion, GeneratorPlugin

from .reader import PhantomReader, PhantomStatsReader
from .utils import PhantomConfig
from .widget import PhantomInfoWidget, PhantomProgressBarWidget
from ..Aggregator import Plugin as AggregatorPlugin
from ..Autostop import Plugin as AutostopPlugin
from ..Console import Plugin as ConsolePlugin

logger = logging.getLogger(__name__)


class Plugin(AbstractPlugin, GeneratorPlugin):
    """     Plugin for running phantom tool    """

    OPTION_CONFIG = "config"
    SECTION = PhantomConfig.SECTION

    def __init__(self, core):
        AbstractPlugin.__init__(self, core)
        self.config = None
        self.process = None

        self.predefined_phout = None
        self.phout_import_mode = False
        self.did_phout_import_try = False

        self.phantom_path = None
        self.eta_file = None
        self.processed_ammo_count = 0
        self.phantom_start_time = time.time()
        self.buffered_seconds = "2"

        self.taskset_affinity = None
        self.cpu_count = mp.cpu_count()

        self.phantom = None
        self.cached_info = None
        self.phantom_stderr = None

        self.exclude_markers = []

        self.enum_ammo = False

    @staticmethod
    def get_key():
        return __file__

    def get_available_options(self):
        opts = [
            "phantom_path", "buffered_seconds", "exclude_markers", "affinity"
        ]
        opts += [PhantomConfig.OPTION_PHOUT, self.OPTION_CONFIG]
        opts += PhantomConfig.get_available_options()
        return opts

    def configure(self):
        # plugin part
        self.config = self.get_option(self.OPTION_CONFIG, '')
        self.phantom_path = self.get_option("phantom_path", 'phantom')
        self.enum_ammo = self.get_option("enum_ammo", False)
        self.buffered_seconds = int(
            self.get_option("buffered_seconds", self.buffered_seconds))
        self.exclude_markers = set(
            filter((lambda marker: marker != ''),
                   self.get_option('exclude_markers', []).split(' ')))
        self.taskset_affinity = self.get_option('affinity', '')

        try:
            autostop = self.core.get_plugin_of_type(AutostopPlugin)
            autostop.add_criterion_class(UsedInstancesCriterion)
        except KeyError:
            logger.debug(
                "No autostop plugin found, not adding instances criterion")

        self.predefined_phout = self.get_option(PhantomConfig.OPTION_PHOUT, '')
        if not self.get_option(
                self.OPTION_CONFIG, '') and self.predefined_phout:
            self.phout_import_mode = True

        if not self.config and not self.phout_import_mode:
            self.phantom = PhantomConfig(self.core)
            self.phantom.read_config()

    def prepare_test(self):
        aggregator = self.core.job.aggregator_plugin

        if not self.config and not self.phout_import_mode:

            # generate config
            self.config = self.phantom.compose_config()
            args = [self.phantom_path, 'check', self.config]

            try:
                result = execute(args, catch_out=True)
            except OSError:
                raise RuntimeError("Phantom I/O engine is not installed!")

            retcode = result[0]
            if retcode:
                raise RuntimeError(
                    "Config check failed. Subprocess returned code %s" %
                    retcode)
            if result[2]:
                raise RuntimeError(
                    "Subprocess returned message: %s" % result[2])
            reader = PhantomReader(self.phantom.phout_file)
            logger.debug(
                "Linking sample reader to aggregator."
                " Reading samples from %s", self.phantom.phout_file)

            logger.debug(
                "Linking stats reader to aggregator."
                " Reading stats from %s", self.phantom.stat_log)
        else:
            reader = PhantomReader(self.predefined_phout)
            logger.debug(
                "Linking sample reader to aggregator."
                " Reading samples from %s", self.predefined_phout)
        if aggregator:
            aggregator.reader = reader
            info = self.phantom.get_info()
            aggregator.stats_reader = PhantomStatsReader(
                self.phantom.stat_log, info)

            aggregator.add_result_listener(self)
        try:
            console = self.core.get_plugin_of_type(ConsolePlugin)
        except Exception as ex:
            logger.debug("Console not found: %s", ex)
            console = None

        self.core.job.phantom_info = self.phantom.get_info()

        if console and aggregator:
            widget = PhantomProgressBarWidget(self)
            console.add_info_widget(widget)
            aggregator.add_result_listener(widget)

            widget = PhantomInfoWidget(self)
            console.add_info_widget(widget)
            aggregator = self.core.get_plugin_of_type(AggregatorPlugin)
            aggregator.add_result_listener(widget)

    def start_test(self):
        if not self.phout_import_mode:
            args = [self.phantom_path, 'run', self.config]
            logger.debug(
                "Starting %s with arguments: %s", self.phantom_path, args)
            if self.taskset_affinity != '':
                args = [
                    self.core.taskset_path, '-c', self.taskset_affinity
                ] + args
                logger.debug(
                    "Enabling taskset for phantom with affinity: %s,"
                    " cores count: %d", self.taskset_affinity, self.cpu_count)
            self.phantom_start_time = time.time()
            phantom_stderr_file = self.core.mkstemp(
                ".log", "phantom_stdout_stderr_")
            self.core.add_artifact_file(phantom_stderr_file)
            self.phantom_stderr = open(phantom_stderr_file, 'w')
            self.process = subprocess.Popen(
                args,
                stderr=self.phantom_stderr,
                stdout=self.phantom_stderr,
                close_fds=True)
        else:
            if not os.path.exists(self.predefined_phout):
                raise RuntimeError(
                    "Phout file not exists for import: %s" %
                    self.predefined_phout)
            logger.warn(
                "Will import phout file instead of running phantom: %s",
                self.predefined_phout)

    def is_test_finished(self):
        if not self.phout_import_mode:
            retcode = self.process.poll()
            if retcode is not None:
                logger.info("Phantom done its work with exit code: %s", retcode)
                return abs(retcode)
            else:
                info = self.get_info()
                if info:
                    eta = int(info.duration) - (
                        int(time.time()) - int(self.phantom_start_time))
                    self.publish('eta', eta)
                return -1
        else:
            if not self.processed_ammo_count or self.did_phout_import_try != self.processed_ammo_count:
                self.did_phout_import_try = self.processed_ammo_count
                return -1
            else:
                return 0

    def end_test(self, retcode):
        if self.process and self.process.poll() is None:
            logger.warn(
                "Terminating phantom process with PID %s", self.process.pid)
            self.process.terminate()
            if self.process:
                self.process.communicate()
        else:
            logger.debug("Seems phantom finished OK")
        if self.phantom_stderr:
            self.phantom_stderr.close()
        return retcode

    def post_process(self, retcode):
        if not retcode:
            info = self.get_info()
            if info and info.ammo_count != self.processed_ammo_count:
                logger.warning(
                    "Planned ammo count %s differs from processed %s",
                    info.ammo_count, self.processed_ammo_count)
        return retcode

    def on_aggregated_data(self, data, stat):
        self.processed_ammo_count += data["overall"]["interval_real"]["len"]
        logger.debug("Processed ammo count: %s/", self.processed_ammo_count)

    def get_info(self):
        """ returns info object """
        if not self.cached_info:
            if not self.phantom:
                return None
            self.cached_info = self.phantom.get_info()
        return self.cached_info


class UsedInstancesCriterion(AbstractCriterion):
    """
    Autostop criterion, based on active instances count
    """
    RC_INST = 24

    @staticmethod
    def get_type_string():
        return 'instances'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        self.autostop = autostop
        self.threads_limit = 1

        level_str = param_str.split(',')[0].strip()
        if level_str[-1:] == '%':
            self.level = float(level_str[:-1]) / 100
            self.is_relative = True
        else:
            self.level = int(level_str)
            self.is_relative = False
        self.seconds_limit = expand_to_seconds(param_str.split(',')[1])

        try:
            phantom = autostop.core.get_plugin_of_type(Plugin)
            info = phantom.get_info()
            if info:
                self.threads_limit = info.instances
            if not self.threads_limit:
                raise ValueError(
                    "Cannot create 'instances' criterion"
                    " with zero instances limit")
        except KeyError:
            logger.warning("No phantom module, 'instances' autostop disabled")

    def notify(self, data, stat):
        threads = stat["metrics"]["instances"]
        if self.is_relative:
            threads = float(threads) / self.threads_limit
        if threads > self.level:
            if not self.seconds_count:
                self.cause_second = (data, stat)

            logger.debug(self.explain())

            self.seconds_count += 1
            self.autostop.add_counting(self)
            if self.seconds_count >= self.seconds_limit:
                return True
        else:
            self.seconds_count = 0

        return False

    def get_rc(self):
        return self.RC_INST

    def get_level_str(self):
        """
        String value for instances level
        """
        if self.is_relative:
            level_str = str(100 * self.level) + "%"
        else:
            level_str = self.level
        return level_str

    def explain(self):
        items = (
            self.get_level_str(), self.seconds_count,
            self.cause_second[0].get('ts'))
        return (
            "Testing threads (instances) utilization"
            " higher than %s for %ss, since %s" % items)

    def widget_explain(self):
        items = (self.get_level_str(), self.seconds_count, self.seconds_limit)
        return "Instances >%s for %s/%ss" % items, float(
            self.seconds_count) / self.seconds_limit



'''
Common utilities
'''
import os
import socket
import threading as th
import traceback

import http.client
import logging
import errno
import itertools
import re
import select
import shlex
import psutil
import subprocess
import argparse
from paramiko import SSHClient, AutoAddPolicy

logger = logging.getLogger(__name__)


class Drain(th.Thread):
    """
    Drain a generator to a destination that answers to put(), in a thread
    """

    def __init__(self, source, destination):
        super(Drain, self).__init__()
        self.source = source
        self.destination = destination
        self._interrupted = th.Event()

    def run(self):
        for item in self.source:
            self.destination.put(item)
            if self._interrupted.is_set():
                break

    def close(self):
        self._interrupted.set()


class SecuredShell(object):
    def __init__(self, host, port, username, timeout):
        self.host = host
        self.port = port
        self.username = username
        self.timeout = timeout

    def connect(self):
        logger.debug(
            "Opening SSH connection to {host}:{port}".format(
                host=self.host, port=self.port))
        client = SSHClient()
        client.load_system_host_keys()
        client.set_missing_host_key_policy(AutoAddPolicy())

        try:
            client.connect(
                self.host,
                port=self.port,
                username=self.username,
                timeout=self.timeout, )
        except ValueError as e:
            logger.error(e)
            logger.warning(
                """
Patching Crypto.Cipher.AES.new and making another attempt.

See here for the details:
http://uucode.com/blog/2015/02/20/workaround-for-ctr-mode-needs-counter-parameter-not-iv/
            """)
            client.close()
            import Crypto.Cipher.AES
            orig_new = Crypto.Cipher.AES.new

            def fixed_AES_new(key, *ls):
                if Crypto.Cipher.AES.MODE_CTR == ls[0]:
                    ls = list(ls)
                    ls[1] = ''
                return orig_new(key, *ls)

            Crypto.Cipher.AES.new = fixed_AES_new
            client.connect(
                self.host,
                port=self.port,
                username=self.username,
                timeout=self.timeout, )
        return client

    def execute(self, cmd):
        logger.info("Execute on %s: %s", self.host, cmd)
        with self.connect() as client:
            _, stdout, stderr = client.exec_command(cmd)
            output = stdout.read()
            errors = stderr.read()
            err_code = stdout.channel.recv_exit_status()
        return output, errors, err_code

    def rm(self, path):
        return self.execute("rm -f %s" % path)

    def rm_r(self, path):
        return self.execute("rm -rf %s" % path)

    def mkdir(self, path):
        return self.execute("mkdir -p %s" % path)

    def send_file(self, local_path, remote_path):
        logger.info(
            "Sending [{local}] to {host}:[{remote}]".format(
                local=local_path, host=self.host, remote=remote_path))
        with self.connect() as client, client.open_sftp() as sftp:
            result = sftp.put(local_path, remote_path)
        return result

    def get_file(self, remote_path, local_path):
        logger.info(
            "Receiving from {host}:[{remote}] to [{local}]".format(
                local=local_path, host=self.host, remote=remote_path))
        with self.connect() as client, client.open_sftp() as sftp:
            result = sftp.get(remote_path, local_path)
        return result

    def async_session(self, cmd):
        return AsyncSession(self, cmd)


def check_ssh_connection():
    logging.basicConfig(
        level=logging.DEBUG, format='%(asctime)s %(levelname)s %(message)s')
    logging.getLogger("paramiko.transport").setLevel(logging.DEBUG)

    parser = argparse.ArgumentParser(
        description='Test SSH connection for monitoring.')
    parser.add_argument(
        '-e', '--endpoint', default='example.org', help='which host to try')

    parser.add_argument(
        '-u', '--username', default=os.getlogin(), help='SSH username')

    parser.add_argument('-p', '--port', default=22, type=int, help='SSH port')
    args = parser.parse_args()
    logging.info(
        "Checking SSH to %s@%s:%d", args.username, args.endpoint, args.port)
    ssh = SecuredShell(args.endpoint, args.port, args.username, 10)
    print(ssh.execute("ls -l"))


class AsyncSession(object):
    def __init__(self, ssh, cmd):
        self.client = ssh.connect()
        self.session = self.client.get_transport().open_session()
        self.session.get_pty()
        self.session.exec_command(cmd)

    def send(self, data):
        self.session.send(data)

    def close(self):
        self.session.close()
        self.client.close()

    def finished(self):
        return self.session.exit_status_ready()

    def read_maybe(self):
        if self.session.recv_ready():
            return self.session.recv(4096)
        else:
            return None


# HTTP codes
HTTP = http.client.responses

# Extended list of HTTP status codes(WEBdav etc.)
# HTTP://en.wikipedia.org/wiki/List_of_HTTP_status_codes
WEBDAV = {
    102: 'Processing',
    103: 'Checkpoint',
    122: 'Request-URI too long',
    207: 'Multi-Status',
    226: 'IM Used',
    308: 'Resume Incomplete',
    418: 'I\'m a teapot',
    422: 'Unprocessable Entity',
    423: 'Locked',
    424: 'Failed Dependency',
    425: 'Unordered Collection',
    426: 'Upgrade Required',
    444: 'No Response',
    449: 'Retry With',
    450: 'Blocked by Windows Parental Controls',
    499: 'Client Closed Request',
    506: 'Variant Also Negotiates',
    507: 'Insufficient Storage',
    509: 'Bandwidth Limit Exceeded',
    510: 'Not Extended',
    598: 'network read timeout error',
    599: 'network connect timeout error',
    999: 'Common Failure',
}
HTTP.update(WEBDAV)

# NET codes
NET = {
    0: "Success",
    1: "Operation not permitted",
    2: "No such file or directory",
    3: "No such process",
    4: "Interrupted system call",
    5: "Input/output error",
    6: "No such device or address",
    7: "Argument list too long",
    8: "Exec format error",
    9: "Bad file descriptor",
    10: "No child processes",
    11: "Resource temporarily unavailable",
    12: "Cannot allocate memory",
    13: "Permission denied",
    14: "Bad address",
    15: "Block device required",
    16: "Device or resource busy",
    17: "File exists",
    18: "Invalid cross-device link",
    19: "No such device",
    20: "Not a directory",
    21: "Is a directory",
    22: "Invalid argument",
    23: "Too many open files in system",
    24: "Too many open files",
    25: "Inappropriate ioctl for device",
    26: "Text file busy",
    27: "File too large",
    28: "No space left on device",
    29: "Illegal seek",
    30: "Read-only file system",
    31: "Too many links",
    32: "Broken pipe",
    33: "Numerical argument out of domain",
    34: "Numerical result out of range",
    35: "Resource deadlock avoided",
    36: "File name too long",
    37: "No locks available",
    38: "Function not implemented",
    39: "Directory not empty",
    40: "Too many levels of symbolic links",
    41: "Unknown error 41",
    42: "No message of desired type",
    43: "Identifier removed",
    44: "Channel number out of range",
    45: "Level 2 not synchronized",
    46: "Level 3 halted",
    47: "Level 3 reset",
    48: "Link number out of range",
    49: "Protocol driver not attached",
    50: "No CSI structure available",
    51: "Level 2 halted",
    52: "Invalid exchange",
    53: "Invalid request descriptor",
    54: "Exchange full",
    55: "No anode",
    56: "Invalid request code",
    57: "Invalid slot",
    58: "Unknown error 58",
    59: "Bad font file format",
    60: "Device not a stream",
    61: "No data available",
    62: "Timer expired",
    63: "Out of streams resources",
    64: "Machine is not on the network",
    65: "Package not installed",
    66: "Object is remote",
    67: "Link has been severed",
    68: "Advertise error",
    69: "Srmount error",
    70: "Communication error on send",
    71: "Protocol error",
    72: "Multihop attempted",
    73: "RFS specific error",
    74: "Bad message",
    75: "Value too large for defined data type",
    76: "Name not unique on network",
    77: "File descriptor in bad state",
    78: "Remote address changed",
    79: "Can not access a needed shared library",
    80: "Accessing a corrupted shared library",
    81: ".lib section in a.out corrupted",
    82: "Attempting to link in too many shared libraries",
    83: "Cannot exec a shared library directly",
    84: "Invalid or incomplete multibyte or wide character",
    85: "Interrupted system call should be restarted",
    86: "Streams pipe error",
    87: "Too many users",
    88: "Socket operation on non-socket",
    89: "Destination address required",
    90: "Message too long",
    91: "Protocol wrong type for socket",
    92: "Protocol not available",
    93: "Protocol not supported",
    94: "Socket type not supported",
    95: "Operation not supported",
    96: "Protocol family not supported",
    97: "Address family not supported by protocol",
    98: "Address already in use",
    99: "Cannot assign requested address",
    100: "Network is down",
    101: "Network is unreachable",
    102: "Network dropped connection on reset",
    103: "Software caused connection abort",
    104: "Connection reset by peer",
    105: "No buffer space available",
    106: "Transport endpoint is already connected",
    107: "Transport endpoint is not connected",
    108: "Cannot send after transport endpoint shutdown",
    109: "Too many references: cannot splice",
    110: "Connection timed out",
    111: "Connection refused",
    112: "Host is down",
    113: "No route to host",
    114: "Operation already in progress",
    115: "Operation now in progress",
    116: "Stale NFS file handle",
    117: "Structure needs cleaning",
    118: "Not a XENIX named type file",
    119: "No XENIX semaphores available",
    120: "Is a named type file",
    121: "Remote I/O error",
    122: "Disk quota exceeded",
    123: "No medium found",
    124: "Wrong medium type",
    125: "Operation canceled",
    126: "Required key not available",
    127: "Key has expired",
    128: "Key has been revoked",
    129: "Key was rejected by service",
    130: "Owner died",
    131: "State not recoverable",
    999: 'Common Failure',
}


def log_stdout_stderr(log, stdout, stderr, comment=""):
    """
    This function polls stdout and stderr streams and writes their contents
    to log
    """
    readable = select.select([stdout], [], [], 0)[0]
    if stderr:
        exceptional = select.select([stderr], [], [], 0)[0]
    else:
        exceptional = []

    log.debug("Selected: %s, %s", readable, exceptional)

    for handle in readable:
        line = handle.read()
        readable.remove(handle)
        if line:
            log.debug("%s stdout: %s", comment, line.strip())

    for handle in exceptional:
        line = handle.read()
        exceptional.remove(handle)
        if line:
            log.warn("%s stderr: %s", comment, line.strip())


def expand_to_milliseconds(str_time):
    """
    converts 1d2s into milliseconds
    """
    return expand_time(str_time, 'ms', 1000)


def expand_to_seconds(str_time):
    """
    converts 1d2s into seconds
    """
    return expand_time(str_time, 's', 1)


def expand_time(str_time, default_unit='s', multiplier=1):
    """
    helper for above functions
    """
    parser = re.compile('(\d+)([a-zA-Z]*)')
    parts = parser.findall(str_time)
    result = 0.0
    for value, unit in parts:
        value = int(value)
        unit = unit.lower()
        if unit == '':
            unit = default_unit

        if unit == 'ms':
            result += value * 0.001
            continue
        elif unit == 's':
            result += value
            continue
        elif unit == 'm':
            result += value * 60
            continue
        elif unit == 'h':
            result += value * 60 * 60
            continue
        elif unit == 'd':
            result += value * 60 * 60 * 24
            continue
        elif unit == 'w':
            result += value * 60 * 60 * 24 * 7
            continue
        else:
            raise ValueError(
                "String contains unsupported unit %s: %s" % (unit, str_time))
    return int(result * multiplier)


def pid_exists(pid):
    """Check whether pid exists in the current process table."""
    if pid < 0:
        return False
    try:
        os.kill(pid, 0)
    except OSError as exc:
        logging.debug("No process[%s]: %s", exc.errno, exc)
        return exc.errno == errno.EPERM
    else:
        p = psutil.Process(pid)
        return p.status != psutil.STATUS_ZOMBIE


def execute(cmd, shell=False, poll_period=1.0, catch_out=False):
    """
    Wrapper for Popen
    """
    log = logging.getLogger(__name__)
    log.debug("Starting: %s", cmd)

    stdout = ""
    stderr = ""

    if not shell and isinstance(cmd, basestring):
        cmd = shlex.split(cmd)

    if catch_out:
        process = subprocess.Popen(
            cmd,
            shell=shell,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            close_fds=True)
    else:
        process = subprocess.Popen(cmd, shell=shell, close_fds=True)

    stdout, stderr = process.communicate()
    if stderr:
        log.error("There were errors:\n%s", stderr)

    if stdout:
        log.debug("Process output:\n%s", stdout)
    returncode = process.returncode
    log.debug("Process exit code: %s", returncode)
    return returncode, stdout, stderr


def splitstring(string):
    """
    >>> string = 'apple orange "banana tree" green'
    >>> splitstring(string)
    ['apple', 'orange', 'green', '"banana tree"']
    """
    patt = re.compile(r'"[\w ]+"')
    if patt.search(string):
        quoted_item = patt.search(string).group()
        newstring = patt.sub('', string)
        return newstring.split() + [quoted_item]
    else:
        return string.split()


def pairs(lst):
    """
    Iterate over pairs in the list
    """
    return itertools.izip(lst[::2], lst[1::2])


def update_status(status, multi_key, value):
    if len(multi_key) > 1:
        update_status(status.setdefault(multi_key[0], {}), multi_key[1:], value)
    else:
        status[multi_key[0]] = value


class AddressWizard:
    def __init__(self):
        self.lookup_fn = socket.getaddrinfo
        self.socket_class = socket.socket

    def resolve(self, address_str, do_test=False, explicit_port=False):
        """

        :param address_str:
        :return: tuple of boolean, string, int - isIPv6, resolved_ip, port (may be null), extracted_address
        """

        if not address_str:
            raise RuntimeError("Mandatory option was not specified: address")

        logger.debug("Trying to resolve address string: %s", address_str)

        port = None

        braceport_pat = "^\[([^]]+)\]:(\d+)$"
        braceonly_pat = "^\[([^]]+)\]$"
        if re.match(braceport_pat, address_str):
            logger.debug("Braces and port present")
            match = re.match(braceport_pat, address_str)
            logger.debug("Match: %s %s ", match.group(1), match.group(2))
            address_str, port = match.group(1), match.group(2)
        elif re.match(braceonly_pat, address_str):
            logger.debug("Braces only present")
            match = re.match(braceonly_pat, address_str)
            logger.debug("Match: %s", match.group(1))
            address_str = match.group(1)
        else:
            logger.debug("Parsing port")
            parts = address_str.split(":")
            if len(parts) <= 2:  # otherwise it is v6 address
                address_str = parts[0]
                if len(parts) == 2:
                    port = int(parts[1])
        if port is not None:
            port = int(port)
        try:
            resolved = self.lookup_fn(address_str, port)
            logger.debug("Lookup result: %s", resolved)
        except Exception as exc:
            logger.debug(
                "Exception trying to resolve hostname %s : %s", address_str,
                traceback.format_exc(exc))
            msg = "Failed to resolve hostname: %s. Error: %s"
            raise RuntimeError(msg % (address_str, exc))

        for (family, socktype, proto, canonname, sockaddr) in resolved:
            is_v6 = family == socket.AF_INET6
            parsed_ip, port = sockaddr[0], sockaddr[1]

            if explicit_port:
                logger.warn(
                    "Using phantom.port option is deprecated. Use phantom.address=[address]:port instead"
                )
                port = int(explicit_port)
            elif not port:
                port = 80

            if do_test:
                try:
                    self.__test(family, (parsed_ip, port))
                except RuntimeError as exc:
                    logger.warn(
                        "Failed TCP connection test using [%s]:%s", parsed_ip,
                        port)
                    continue

            return is_v6, parsed_ip, int(port), address_str

        msg = "All connection attempts failed for %s, use phantom.connection_test=0 to disable it"
        raise RuntimeError(msg % address_str)

    def __test(self, af, sa):
        test_sock = self.socket_class(af)
        try:
            test_sock.settimeout(5)
            test_sock.connect(sa)
        except Exception as exc:
            logger.debug(
                "Exception on connect attempt [%s]:%s : %s", sa[0], sa[1],
                traceback.format_exc(exc))
            msg = "TCP Connection test failed for [%s]:%s, use phantom.connection_test=0 to disable it"
            raise RuntimeError(msg % (sa[0], sa[1]))
        finally:
            test_sock.close()


class Chopper(object):
    def __init__(self, source):
        self.source = source

    def __iter__(self):
        for chunk in self.source:
            for item in chunk:
                yield item
